'''
mesoSPIM Stage classes
======================
'''
import time
import logging

logger = logging.getLogger(__name__)

from PyQt5 import QtCore

from .mesoSPIM_State import mesoSPIM_StateSingleton


class mesoSPIM_Stage(QtCore.QObject):
    '''
    DemoStage for a mesoSPIM microscope

    It is expected that the parent class has the following signals:
        sig_move_relative = pyqtSignal(dict)
        sig_move_relative_and_wait_until_done = pyqtSignal(dict)
        sig_move_absolute = pyqtSignal(dict)
        sig_move_absolute_and_wait_until_done = pyqtSignal(dict)
        sig_zero = pyqtSignal(list)
        sig_unzero = pyqtSignal(list)
        sig_stop_movement = pyqtSignal()
        sig_mark_rotation_position = pyqtSignal()

    Also contains a QTimer that regularily sends position updates, e.g
    during the execution of movements.
    '''

    sig_position = QtCore.pyqtSignal(dict)
    sig_status_message = QtCore.pyqtSignal(str)

    sig_pause = QtCore.pyqtSignal(bool)

    def __init__(self, parent = None):
        super().__init__()
        self.parent = parent
        self.cfg = parent.cfg

        # self.state = mesoSPIM_StateSingleton()

        ''' The movement signals are emitted by the mesoSPIM_Core, which in turn
        instantiates the mesoSPIM_Serial thread.

        Therefore, the signals are emitted by the parent of the parent, which
        is slightly confusing and dirty.
        '''

        self.parent.sig_stop_movement.connect(self.stop)
        self.parent.sig_zero_axes.connect(self.zero_axes)
        self.parent.sig_unzero_axes.connect(self.unzero_axes)
        self.parent.sig_load_sample.connect(self.load_sample)
        self.parent.sig_unload_sample.connect(self.unload_sample)
        self.parent.sig_mark_rotation_position.connect(self.mark_rotation_position)

        self.pos_timer = QtCore.QTimer(self)
        self.pos_timer.timeout.connect(self.report_position)
        self.pos_timer.start(100)

        '''Initial setting of all positions

        self.x_pos, self.y_pos etc are the true axis positions, no matter whether
        the stages are zeroed or not.
        '''
        self.x_pos = 0
        self.y_pos = 0
        self.z_pos = 0
        self.f_pos = 0
        self.theta_pos = 0

        '''Internal (software) positions'''
        self.int_x_pos = 0
        self.int_y_pos = 0
        self.int_z_pos = 0
        self.int_f_pos = 0
        self.int_theta_pos = 0

        '''Create offsets

        It should be:

        int_x_pos = x_pos + int_x_pos_offset
        self.int_x_pos = self.x_pos + self.int_x_pos_offset

        OR

        x_pos = int_x_pos - int_x_pos_offset
        self.x_pos = self.int_x_pos - self.int_x_pos_offset

        '''
        self.int_x_pos_offset = 0
        self.int_y_pos_offset = 0
        self.int_z_pos_offset = 0
        self.int_f_pos_offset = 0
        self.int_theta_pos_offset = 0

        '''
        Setting movement limits: currently hardcoded

        Open question: Should these be in mm or microns?
        Answer: Microns for now....
        '''
        self.x_max = self.cfg.stage_parameters['x_max']
        self.x_min = self.cfg.stage_parameters['x_min']
        self.y_max = self.cfg.stage_parameters['y_max']
        self.y_min = self.cfg.stage_parameters['y_min']
        self.z_max = self.cfg.stage_parameters['z_max']
        self.z_min = self.cfg.stage_parameters['z_min']
        self.f_max = self.cfg.stage_parameters['f_max']
        self.f_min = self.cfg.stage_parameters['f_min']
        self.theta_max = self.cfg.stage_parameters['theta_max']
        self.theta_min = self.cfg.stage_parameters['theta_min']
        self.x_rot_position = self.cfg.stage_parameters['x_rot_position']
        self.y_rot_position = self.cfg.stage_parameters['y_rot_position']
        self.z_rot_position = self.cfg.stage_parameters['z_rot_position']

    def create_position_dict(self):
        self.position_dict = {'x_pos': self.x_pos,
                              'y_pos': self.y_pos,
                              'z_pos': self.z_pos,
                              'f_pos': self.f_pos,
                              'theta_pos': self.theta_pos,
                              }

    def create_internal_position_dict(self):
        self.int_position_dict = {'x_pos': self.int_x_pos,
                                  'y_pos': self.int_y_pos,
                                  'z_pos': self.int_z_pos,
                                  'f_pos': self.int_f_pos,
                                  'theta_pos': self.int_theta_pos,
                                  }

    def report_position(self):
        self.create_position_dict()

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        # self.state['position'] = self.int_position_dict

        self.sig_position.emit(self.int_position_dict)

    # @QtCore.pyqtSlot(dict)
    def move_relative(self, dict, wait_until_done=False):
        ''' Move relative method '''
        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                self.x_pos = self.x_pos + x_rel
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                self.y_pos = self.y_pos + y_rel
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                self.z_pos = self.z_pos + z_rel
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                self.theta_pos = self.theta_pos + theta_rel
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                self.f_pos = self.f_pos + f_rel
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if wait_until_done == True:
            time.sleep(0.02)

    # @QtCore.pyqtSlot(dict)
    def move_absolute(self, dict, wait_until_done=False):
        ''' Move absolute method '''

        if 'x_abs' in dict:
            x_abs = dict['x_abs']
            x_abs = x_abs - self.int_x_pos_offset
            if self.x_min < x_abs and self.x_max > x_abs:
                self.x_pos = x_abs
            else:
                self.sig_status_message.emit('Absolute movement stopped: X Motion limit would be reached!')

        if 'y_abs' in dict:
            y_abs = dict['y_abs']
            y_abs = y_abs - self.int_y_pos_offset
            if self.y_min < y_abs and self.y_max > y_abs:
                self.y_pos = y_abs
            else:
                self.sig_status_message.emit('Absolute movement stopped: Y Motion limit would be reached!')

        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs and self.z_max > z_abs:
                self.z_pos = z_abs
            else:
                self.sig_status_message.emit('Absolute movement stopped: Z Motion limit would be reached!')

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs and self.f_max > f_abs:
                self.f_pos = f_abs
            else:
                self.sig_status_message.emit('Absolute movement stopped: F Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs and self.theta_max > theta_abs:
                self.theta_pos = theta_abs
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done == True:
            time.sleep(3)

    @QtCore.pyqtSlot()
    def stop(self):
        self.sig_status_message.emit('Stopped')

    def zero_axes(self, list):
        for axis in list:
            try:
                exec('self.int_' + axis + '_pos_offset = -self.' + axis + '_pos')
            except:
                logger.info('Zeroing of axis: ', axis, 'failed')

    def unzero_axes(self, list):
        for axis in list:
            try:
                exec('self.int_' + axis + '_pos_offset = 0')
            except:
                logger.info('Unzeroing of axis: ', axis, 'failed')

    def load_sample(self):
        self.y_pos = self.cfg.stage_parameters['y_load_position']

    def unload_sample(self):
        self.y_pos = self.cfg.stage_parameters['y_unload_position']

    def mark_rotation_position(self):
        ''' Take the current position and mark it as rotation location '''
        self.x_rot_position = self.x_pos
        self.y_rot_position = self.y_pos
        self.z_rot_position = self.z_pos
        logger.info('Marking new rotation position (absolute coordinates): X: ', self.x_pos, ' Y: ', self.y_pos, ' Z: ',
                    self.z_pos)

    def go_to_rotation_position(self, wait_until_done=False):
        ''' Move to the proper rotation position.
        '''
        print('Going to rotation position: NOT IMPLEMENTED / DEMO MODE')
        logger.info('Going to rotation position: NOT IMPLEMENTED / DEMO MODE')


class mesoSPIM_DemoStage(mesoSPIM_Stage):
    def __init__(self, parent=None):
        super().__init__(parent)


class mesoSPIM_PI_1toN(mesoSPIM_Stage):
    '''
    Configuration with 1 controller connected to N stages, (e.g. C-884, default mesoSPIM V5 setup).

    Note:
    configs as declared in mesoSPIM_config.py:
        stage_parameters = {'stage_type' : 'PI_1controllerNstages',
                            ...
                            }
    pi_parameters = {'controllername' : 'C-884',
                    'stages' : ('L-509.20DG10','L-509.40DG10','L-509.20DG10','M-060.DG','M-406.4PD','NOSTAGE'),
                    'refmode' : ('FRF',),
                    'serialnum' : ('118075764'),
                    }
    '''

    def __init__(self, parent=None):
        super().__init__(parent)
        from pipython import GCSDevice, pitools
        self.pitools = pitools

        ''' Setting up the PI stages '''
        self.pi = self.cfg.pi_parameters
        self.controllername = self.cfg.pi_parameters['controllername']
        self.pi_stages = list(self.cfg.pi_parameters['stages'])
        self.refmode = self.cfg.pi_parameters['refmode']
        self.serialnum = self.cfg.pi_parameters['serialnum']
        self.pidevice = GCSDevice(self.controllername)
        self.pidevice.ConnectUSB(serialnum=self.serialnum)

        ''' PI startup '''
        ''' with refmode enabled: pretty dangerous
        pitools.startup(self.pidevice, stages=self.pi_stages, refmode=self.refmode)
        '''
        pitools.startup(self.pidevice, stages=self.pi_stages)

        ''' Report reference status of all stages '''
        for ii in range(1, len(self.pi_stages) + 1):
            tStage = self.pi_stages[ii - 1]
            if tStage == 'NOSTAGE':
                continue

            tState = self.pidevice.qFRF(ii)
            if tState[ii]:
                msg = 'referenced'
            else:
                msg = '*UNREFERENCED*'

            logger.info("Axis %d (%s) reference status: %s" % (ii, tStage, msg))

        ''' Stage 5 referencing hack '''
        self.pidevice.FRF(5)
        logger.info('M-406 Emergency referencing hack: Waiting for referencing move')
        self.block_till_controller_is_ready()
        logger.info('M-406 Emergency referencing hack done')

    def __del__(self):
        try:
            self.pidevice.unload()
        except:
            pass

    def report_position(self):
        positions = self.pidevice.qPOS(self.pidevice.axes)
        self.x_pos = round(positions['1'] * 1000, 2)
        self.y_pos = round(positions['2'] * 1000, 2)
        self.z_pos = round(positions['3'] * 1000, 2)
        self.f_pos = round(positions['5'] * 1000, 2)
        self.theta_pos = positions['4']

        self.create_position_dict()

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()
        # self.state['position'] = self.int_position_dict
        self.sig_position.emit(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' PI move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel < self.x_max:
                x_rel = x_rel / 1000
                self.pidevice.MVR({1: x_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel < self.y_max:
                y_rel = y_rel / 1000
                self.pidevice.MVR({2: y_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel < self.z_max:
                z_rel = z_rel / 1000
                self.pidevice.MVR({3: z_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel < self.theta_max:
                self.pidevice.MVR({4: theta_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel < self.f_max:
                f_rel = f_rel / 1000
                self.pidevice.MVR({5: f_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if wait_until_done:
            self.pitools.waitontarget(self.pidevice)

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Lots of implementation details in here, should be replaced by a facade

        TODO: Also lots of repeating code.
        TODO: DRY principle violated
        '''

        if 'x_abs' in dict:
            x_abs = dict['x_abs']
            x_abs = x_abs - self.int_x_pos_offset
            if self.x_min < x_abs < self.x_max:
                ''' Conversion to mm and command emission'''
                x_abs = x_abs / 1000
                self.pidevice.MOV({1: x_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: X Motion limit would be reached!')

        if 'y_abs' in dict:
            y_abs = dict['y_abs']
            y_abs = y_abs - self.int_y_pos_offset
            if self.y_min < y_abs < self.y_max:
                ''' Conversion to mm and command emission'''
                y_abs = y_abs / 1000
                self.pidevice.MOV({2: y_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Y Motion limit would be reached!')

        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs < self.z_max:
                ''' Conversion to mm and command emission'''
                z_abs = z_abs / 1000
                self.pidevice.MOV({3: z_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Z Motion limit would be reached!')

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs < self.f_max:
                ''' Conversion to mm and command emission'''
                f_abs = f_abs / 1000
                self.pidevice.MOV({5: f_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: F Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs < self.theta_max:
                ''' No Conversion to mm !!!! and command emission'''
                self.pidevice.MOV({4: theta_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done:
            self.pitools.waitontarget(self.pidevice)

    def stop(self):
        self.pidevice.STP(noraise=True)

    def load_sample(self):
        y_abs = self.cfg.stage_parameters['y_load_position'] / 1000
        self.pidevice.MOV({2: y_abs})

    def unload_sample(self):
        y_abs = self.cfg.stage_parameters['y_unload_position'] / 1000
        self.pidevice.MOV({2: y_abs})

    def go_to_rotation_position(self, wait_until_done=False):
        x_abs = self.x_rot_position / 1000
        y_abs = self.y_rot_position / 1000
        z_abs = self.z_rot_position / 1000

        self.pidevice.MOV({1: x_abs, 2: y_abs, 3: z_abs})

        if wait_until_done:
            self.pitools.waitontarget(self.pidevice)

    def block_till_controller_is_ready(self):
        '''
        Blocks further execution (especially during referencing moves)
        till the PI controller returns ready
        '''
        blockflag = True
        while blockflag:
            if self.pidevice.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)


class mesoSPIM_PI_NtoN(mesoSPIM_Stage):
    '''
    Expects following microscope configuration:
        Sample XYZ movement: Physik Instrumente stage with three L-509-type stepper motor stages and individual C-663 controller.
        F movement: Physik Instrumente C-663 controller and custom stage with stepper motor
        Rotation: not implemented

        All stage controller are of same type and the sample stages work with reference setting.
        Focus stage has reference mode set to off.
            
    Note:
        configs as declared in mesoSPIM_config.py:
        stage_parameters = {'stage_type' : 'PI_NcontrollersNstages',
                            ...
                            }
        pi_parameters = {'axes_names': ('x', 'y', 'z', 'theta', 'f'),
                        'stages': ('L-509.20SD00', 'L-509.40SD00', 'L-509.20SD00', None, 'MESOSPIM_FOCUS'),
                        'controllername': ('C-663', 'C-663', 'C-663', None, 'C-663'),
                        'serialnum': ('**********', '**********', '**********', None, '**********'),
                        'refmode': ('FRF', 'FRF', 'FRF', None, 'RON')
                        }
        make sure that reference points are not in conflict with general microscope setup
        and will not hurt optics under referencing at startup
    '''

    def __init__(self, parent=None):
        super().__init__(parent)
        from pipython import GCSDevice, pitools
        self.pitools = pitools
        self.pi = self.cfg.pi_parameters
        print("Connecting stage drive...")

        # Setting up the stages with separate PI controller.
        # Explicitly set referencing status and get position

        # gather stage devices in VirtualStages class
        class VirtualStages:
            pass

        assert len(self.pi['axes_names']) == len(self.pi['stages']) == len(self.pi['controllername']) \
               == len(self.pi['serialnum']) == len(self.pi['refmode']), \
            "Config file, pi_parameters dictionary: numbers of axes_names, stages, controllername, serialnum, refmode must match "
        self.pi_stages = VirtualStages()
        for axis_name, stage, controller, serialnum, refmode in zip(self.pi['axes_names'], self.pi['stages'],
                                                                    self.pi['controllername'], self.pi['serialnum'],
                                                                    self.pi['refmode']):
            # run stage startup procedure for each axis
            if stage:
                print(f'starting stage {stage}')
                pidevice_ = GCSDevice(controller)
                pidevice_.ConnectUSB(serialnum=serialnum)
                if refmode is None:
                    pitools.startup(pidevice_, stages=stage)
                elif refmode == 'FRF':
                    pitools.startup(pidevice_, stages=stage, refmodes=refmode)
                    pidevice_.FRF(1)
                elif refmode == 'RON':
                    pitools.startup(pidevice_, stages=stage)
                    pidevice_.RON({1: 0})  # set reference mode
                    # activate servo
                    pidevice_.SVO(pidevice_.axes, [True] * len(pidevice_.axes))
                    # print('servo state: {}'.format(pidevice_.qSVO()))
                    # set/get actual position as home position
                    # assumes that starting position is within reasonable distance from optimal focus
                    pidevice_.POS({1: 0.0})
                    pidevice_.DFH(1)
                else:
                    raise ValueError(f"refmode {refmode} is not supported, PI stage {stage} initialization failed")
                print(f'stage {stage} started')
                print('axis {}, referencing mode: {}'.format(axis_name, pidevice_.qRON()))
                self.wait_for_controller(pidevice_)
                print('axis {}, stage {} ready'.format(axis_name, stage))
                setattr(self.pi_stages, ('pidevice_' + axis_name), pidevice_)
            else:
                setattr(self.pi_stages, axis_name, None)

        logger.info('mesoSPIM_PI_NtoN: started')


    def wait_for_controller(self, controller):
        # function used during stage setup
        blockflag = True
        while blockflag:
            if controller.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)


    def __del__(self):
        '''Close the PI connection'''
        try:
            [(getattr(self.pi_stages, ('pidevice_' + axis_name))).unload() for axis_name in self.pi['axes_names'] if
             (hasattr(self.pi_stages, ('pidevice_' + axis_name)))]
        except:
            pass


    def report_position(self):
        '''report stage position'''
        for axis_name in self.pi['axes_names']:
            pidevice_name = 'pidevice_' + str(axis_name)
            if hasattr(self.pi_stages, pidevice_name):
                try:
                    if axis_name is None:
                        pos = 0
                    elif axis_name == 'theta':
                        pos = (getattr(self.pi_stages, pidevice_name)).qPOS(1)[1]
                    else:
                        pos = round((getattr(self.pi_stages, pidevice_name)).qPOS(1)[1] * 1000, 2)
                except:
                    print(f"Failed to report_position for axis_name {axis_name}, pidevice_name {pidevice_name}.")
            else:
                pos = 0

            setattr(self, (axis_name + '_pos'), pos)
            int_pos = pos + getattr(self, ('int_' + axis_name + '_pos_offset'))
            setattr(self, ('int_' + axis_name + '_pos'), int_pos)

        self.create_position_dict()
        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)


    def move_relative(self, move_dict, wait_until_done=False):
        ''' PI move relative method '''        
        for axis_move in move_dict.keys():        
            axis_name = axis_move.split('_')[0]
            move_value = move_dict[axis_move]        
        
            if (hasattr(self.pi_stages, ('pidevice_' + axis_name))):
                if (getattr(self, (axis_name + '_min')) < getattr(self, (axis_name + '_pos')) + move_value) and \
                    (getattr(self, (axis_name + '_max')) > getattr(self, (axis_name + '_pos')) + move_value):
                    if not axis_name=='theta':
                        move_value = move_value/1000
                    (getattr(self.pi_stages, ('pidevice_' + axis_name))).MVR({1 : move_value})
                else:
                    self.sig_status_message.emit('Relative movement stopped: {} Motion limit would be reached!'.format(axis_name))
                if (axis_name == 'f') or (wait_until_done == True):
                    self.pitools.waitontarget(getattr(self.pi_stages, ('pidevice_' + axis_name)))  # focus may be slower than expected


    def move_absolute(self, move_dict, wait_until_done=False):
        ''' PI move absolute method '''
        for axis_move in move_dict.keys():
            axis_name = axis_move.split('_')[0]
            move_value = move_dict[axis_move] 
            move_value = move_value - getattr(self, ('int_' + axis_name + '_pos_offset'))
            
            if (hasattr(self.pi_stages, ('pidevice_' + axis_name))):
                if (getattr(self, (axis_name + '_min')) < move_value) and \
                        (getattr(self, (axis_name + '_max')) > move_value):
                    if not axis_name == 'theta':
                        move_value = move_value / 1000
                    (getattr(self.pi_stages, ('pidevice_' + axis_name))).MOV({1: move_value})
                else:
                    self.sig_status_message.emit(
                        'Absolute movement stopped: {} Motion limit would be reached!'.format(axis_name))
                if (axis_name == 'f') or (wait_until_done == True):
                    self.pitools.waitontarget(getattr(self.pi_stages, ('pidevice_' + axis_name)))  # focus may be slower than expected


    def stop(self):
        '''stop stage movement'''
        [(getattr(self.pi_stages, ('pidevice_' + axis_name))).STP(noraise=True) for axis_name in self.pi['axes_names']
         if (hasattr(self.pi_stages, ('pidevice_' + axis_name)))]


    def load_sample(self):
        '''bring sample to imaging position'''
        axis_name = 'y'
        y_abs = self.cfg.stage_parameters['y_load_position'] / 1000
        (getattr(self.pi_stages, ('pidevice_' + axis_name))).MOV({1: y_abs})


    def unload_sample(self):
        '''lift sample to sample handling position'''
        axis_name = 'y'
        y_abs = self.cfg.stage_parameters['y_unload_position'] / 1000
        (getattr(self.pi_stages, ('pidevice_' + axis_name))).MOV({1: y_abs})


    '''
    # currently not implemented for this microscope configuration
    def go_to_rotation_position(self, wait_until_done=False):
        x_abs = self.x_rot_position/1000
        y_abs = self.y_rot_position/1000
        z_abs = self.z_rot_position/1000

        self.pidevice.MOV({1 : x_abs, 2 : y_abs, 3 : z_abs})

        if wait_until_done == True:
            self.pitools.waitontarget(self.pidevice)
    '''


class mesoSPIM_GalilStages(mesoSPIM_Stage):
    '''

    It is expected that the parent class has the following signals:
        sig_move_relative = pyqtSignal(dict)
        sig_move_relative_and_wait_until_done = pyqtSignal(dict)
        sig_move_absolute = pyqtSignal(dict)
        sig_move_absolute_and_wait_until_done = pyqtSignal(dict)
        sig_zero = pyqtSignal(list)
        sig_unzero = pyqtSignal(list)
        sig_stop_movement = pyqtSignal()
        sig_mark_rotation_position = pyqtSignal()

    Also contains a QTimer that regularily sends position updates, e.g
    during the execution of movements.

    Todo: Rotation handling not implemented!
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        '''
        Galil-specific code
        '''
        from .devices.stages.galil.galilcontrol import StageControlGalil

        self.x_encodercounts_per_um = self.cfg.xyz_galil_parameters['x_encodercounts_per_um']
        self.y_encodercounts_per_um = self.cfg.xyz_galil_parameters['y_encodercounts_per_um']
        self.z_encodercounts_per_um = self.cfg.xyz_galil_parameters['z_encodercounts_per_um']
        self.f_encodercounts_per_um = self.cfg.f_galil_parameters['z_encodercounts_per_um']

        ''' Setting up the Galil stages '''
        self.xyz_stage = StageControlGalil(COMport=self.cfg.xyz_galil_parameters['COMport'],
                                           x_encodercounts_per_um=self.x_encodercounts_per_um,
                                           y_encodercounts_per_um=self.y_encodercounts_per_um,
                                           z_encodercounts_per_um=self.z_encodercounts_per_um)

        self.f_stage = StageControlGalil(COMport=self.cfg.f_galil_parameters['COMport'],
                                         x_encodercounts_per_um=0,
                                         y_encodercounts_per_um=0,
                                         z_encodercounts_per_um=self.f_encodercounts_per_um)
        '''
        print('Galil: ', self.xyz_stage.read_position('x'))
        print('Galil: ', self.xyz_stage.read_position('y'))
        print('Galil: ', self.xyz_stage.read_position('z'))
        '''

    def __del__(self):
        try:
            '''Close the Galil connection'''
            self.xyz_stage.close_stage()
            self.f_stage.close_stage()
        except:
            pass

    def report_position(self):
        self.x_pos = self.xyz_stage.read_position('x')
        self.y_pos = self.xyz_stage.read_position('y')
        self.z_pos = self.xyz_stage.read_position('z')
        self.f_pos = self.f_stage.read_position('z')
        self.theta_pos = 0

        self.create_position_dict()

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' Galil move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                self.xyz_stage.move_relative(xrel=int(x_rel))
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                self.xyz_stage.move_relative(yrel=int(y_rel))
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                self.xyz_stage.move_relative(zrel=int(z_rel))
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                print('No rotation stage attached')
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                self.f_stage.move_relative(zrel=f_rel)
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if wait_until_done == True:
            pass

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Galil move absolute method

        Lots of implementation details in here, should be replaced by a facade

        '''
        # print(dict)

        # if ('x_abs', 'y_abs', 'z_abs', 'f_abs') in dict:
        x_abs = dict['x_abs']
        x_abs = x_abs - self.int_x_pos_offset
        y_abs = dict['y_abs']
        y_abs = y_abs - self.int_y_pos_offset
        z_abs = dict['z_abs']
        z_abs = z_abs - self.int_z_pos_offset
        f_abs = dict['f_abs']
        f_abs = f_abs - self.int_f_pos_offset

        self.xyz_stage.move_absolute(xabs=x_abs, yabs=y_abs, zabs=z_abs)
        self.f_stage.move_absolute(zabs=f_abs)

        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')

    # def stop(self):
    #     # self.pidevice.STP(noraise=True)

    # def load_sample(self):
    #     y_abs = self.cfg.stage_parameters['y_load_position']/1000
    #     # self.pidevice.MOV({2 : y_abs})

    # def unload_sample(self):
    #     y_abs = self.cfg.stage_parameters['y_unload_position']/1000
    #     # self.pidevice.MOV({2 : y_abs})


class mesoSPIM_PI_f_rot_and_Galil_xyz_Stages(mesoSPIM_Stage):
    '''
    Deprecated?
    Todo: Rotation handling not implemented!
    Todo: Rotation axes are hardcoded! (M-605: #5, M-061.PD: #6)
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        # self.state = mesoSPIM_StateSingleton()

        self.pos_timer = QtCore.QTimer(self)
        self.pos_timer.timeout.connect(self.report_position)
        self.pos_timer.start(50)
        '''
        Galil-specific code
        '''
        from .devices.stages.galil.galilcontrol import StageControlGalil

        self.x_encodercounts_per_um = self.cfg.xyz_galil_parameters['x_encodercounts_per_um']
        self.y_encodercounts_per_um = self.cfg.xyz_galil_parameters['y_encodercounts_per_um']
        self.z_encodercounts_per_um = self.cfg.xyz_galil_parameters['z_encodercounts_per_um']

        ''' Setting up the Galil stages '''
        self.xyz_stage = StageControlGalil(self.cfg.xyz_galil_parameters['port'], [self.x_encodercounts_per_um,
                                                                                   self.y_encodercounts_per_um,
                                                                                   self.z_encodercounts_per_um])
        '''
        self.f_stage = StageControlGalil(COMport = self.cfg.f_galil_parameters['COMport'],
                                        x_encodercounts_per_um = 0,
                                        y_encodercounts_per_um = 0,
                                        z_encodercounts_per_um = self.f_encodercounts_per_um)
        '''

        '''
        print('Galil: ', self.xyz_stage.read_position('x'))
        print('Galil: ', self.xyz_stage.read_position('y'))
        print('Galil: ', self.xyz_stage.read_position('z'))
        '''

        ''' PI-specific code '''
        from pipython import GCSDevice, pitools

        self.pitools = pitools

        ''' Setting up the PI stages '''
        self.pi = self.cfg.pi_parameters

        self.controllername = self.cfg.pi_parameters['controllername']
        self.pi_stages = list(self.cfg.pi_parameters['stages'])
        # ('M-112K033','L-406.40DG10','M-112K033','M-116.DG','M-406.4PD','NOSTAGE')
        self.refmode = self.cfg.pi_parameters['refmode']
        # self.serialnum = ('118015439')  # Wyss Geneva
        self.serialnum = self.cfg.pi_parameters['serialnum']  # UZH Irchel H45

        self.pidevice = GCSDevice(self.controllername)
        self.pidevice.ConnectUSB(serialnum=self.serialnum)

        ''' PI startup '''

        ''' with refmode enabled: pretty dangerous
        pitools.startup(self.pidevice, stages=self.pi_stages, refmode=self.refmode)
        '''
        pitools.startup(self.pidevice, stages=self.pi_stages)

        ''' Setting PI velocities '''
        self.pidevice.VEL(self.cfg.pi_parameters['velocity'])

        ''' Stage 5 referencing hack '''
        # print('Referencing status 3: ', self.pidevice.qFRF(3))
        # print('Referencing status 5: ', self.pidevice.qFRF(5))
        self.pidevice.FRF(5)
        print('M-406 Emergency referencing hack: Waiting for referencing move')
        logger.info('M-406 Emergency referencing hack: Waiting for referencing move')
        self.block_till_controller_is_ready()
        print('M-406 Emergency referencing hack done')
        logger.info('M-406 Emergency referencing hack done')
        # print('Again: Referencing status 3: ', self.pidevice.qFRF(3))
        # print('Again: Referencing status 5: ', self.pidevice.qFRF(5))

    def __del__(self):
        try:
            '''Close the Galil connection'''
            self.xyz_stage.close()
            self.f_stage.close_stage()
        except:
            pass

    def report_position(self):
        positions = self.pidevice.qPOS(self.pidevice.axes)

        '''
        Ugly workaround to deal with non-responding stage 
        position reports: Do not update positions in 
        exceptional circumstances. 
        '''
        self.x_pos, self.y_pos, self.z_pos = self.xyz_stage.read_position()
        self.f_pos = round(positions['5'] * 1000, 2)
        self.theta_pos = positions['6']

        self.create_position_dict()

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)
        # print(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' Galil move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        xyz_motion_dict = {}

        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                xyz_motion_dict.update({1: int(x_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                xyz_motion_dict.update({2: int(y_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                xyz_motion_dict.update({3: int(z_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if xyz_motion_dict != {}:
            self.xyz_stage.move_relative(xyz_motion_dict)

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                self.pidevice.MVR({6: theta_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                f_rel = f_rel / 1000
                self.pidevice.MVR({5: f_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')
            self.pitools.waitontarget(self.pidevice)

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Galil move absolute method

        Lots of implementation details in here, should be replaced by a facade

        '''
        xyz_motion_dict = {}

        if 'x_abs' or 'y_abs' or 'z_abs' in dict:
            if 'x_abs' in dict:
                x_abs = dict['x_abs']
                x_abs = x_abs - self.int_x_pos_offset
                xyz_motion_dict.update({1: x_abs})

            if 'y_abs' in dict:
                y_abs = dict['y_abs']
                y_abs = y_abs - self.int_y_pos_offset
                xyz_motion_dict.update({2: y_abs})

            if 'z_abs' in dict:
                z_abs = dict['z_abs']
                z_abs = z_abs - self.int_z_pos_offset
                xyz_motion_dict.update({3: z_abs})

        if xyz_motion_dict != {}:
            self.xyz_stage.move_absolute(xyz_motion_dict)

        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs and self.f_max > f_abs:
                ''' Conversion to mm and command emission'''
                f_abs = f_abs / 1000
                self.pidevice.MOV({5: f_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: F Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs and self.theta_max > theta_abs:
                ''' No Conversion to mm !!!! and command emission'''
                self.pidevice.MOV({6: theta_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done == True:
            self.pitools.waitontarget(self.pidevice)

    def stop(self):
        self.xyz_stage.stop(restart_programs=True)
        self.pidevice.STP(noraise=True)

    def load_sample(self):
        self.xyz_stage.move_absolute(
            {1: self.int_x_pos, 2: self.cfg.stage_parameters['y_load_position'], 3: self.int_z_pos})

    def unload_sample(self):
        self.xyz_stage.move_absolute(
            {1: self.int_x_pos, 2: self.cfg.stage_parameters['y_unload_position'], 3: self.int_z_pos})

    def go_to_rotation_position(self, wait_until_done=False):
        self.xyz_stage.move_absolute({1: self.x_rot_position, 2: self.y_rot_position, 3: self.z_rot_position})
        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')

    def block_till_controller_is_ready(self):
        '''
        Blocks further execution (especially during referencing moves)
        till the PI controller returns ready
        '''
        blockflag = True
        while blockflag:
            if self.pidevice.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)

    def execute_program(self):
        '''Executes program stored on the Galil controller'''
        self.xyz_stage.execute_program()


class mesoSPIM_PI_rot_and_Galil_xyzf_Stages(mesoSPIM_Stage):
    '''
    Expects following microscope configuration:
    
    Sample XYZ movement: Galil controller with 3 axes 
    F movement: Second Galil controller with a single axis 
    Rotation: PI C-863 mercury controller

    It is expected that the parent class has the following signals:
        sig_move_relative = pyqtSignal(dict)
        sig_move_relative_and_wait_until_done = pyqtSignal(dict)
        sig_move_absolute = pyqtSignal(dict)
        sig_move_absolute_and_wait_until_done = pyqtSignal(dict)
        sig_zero = pyqtSignal(list)
        sig_unzero = pyqtSignal(list)
        sig_stop_movement = pyqtSignal()
        sig_mark_rotation_position = pyqtSignal()

    Also contains a QTimer that regularily sends position updates, e.g
    during the execution of movements.
   
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        # self.state = mesoSPIM_StateSingleton()

        self.pos_timer = QtCore.QTimer(self)
        self.pos_timer.timeout.connect(self.report_position)
        self.pos_timer.start(50)
        '''
        Galil-specific code
        '''
        from .devices.stages.galil.galilcontrol import StageControlGalil

        self.x_encodercounts_per_um = self.cfg.xyz_galil_parameters['x_encodercounts_per_um']
        self.y_encodercounts_per_um = self.cfg.xyz_galil_parameters['y_encodercounts_per_um']
        self.z_encodercounts_per_um = self.cfg.xyz_galil_parameters['z_encodercounts_per_um']
        self.f_encodercounts_per_um = self.cfg.f_galil_parameters['f_encodercounts_per_um']

        ''' Setting up the Galil stages: XYZ '''
        self.xyz_stage = StageControlGalil(self.cfg.xyz_galil_parameters['port'], [self.x_encodercounts_per_um,
                                                                                   self.y_encodercounts_per_um,
                                                                                   self.z_encodercounts_per_um])

        ''' Setting up the Galil stages: F with two dummy axes.'''
        self.f_stage = StageControlGalil(self.cfg.f_galil_parameters['port'], [self.x_encodercounts_per_um,
                                                                               self.y_encodercounts_per_um,
                                                                               self.f_encodercounts_per_um])
        '''
        self.f_stage = StageControlGalil(COMport = self.cfg.f_galil_parameters['COMport'],
                                        x_encodercounts_per_um = 0,
                                        y_encodercounts_per_um = 0,
                                        z_encodercounts_per_um = self.f_encodercounts_per_um)
        '''

        '''
        print('Galil: ', self.xyz_stage.read_position('x'))
        print('Galil: ', self.xyz_stage.read_position('y'))
        print('Galil: ', self.xyz_stage.read_position('z'))
        '''

        ''' PI-specific code '''
        from pipython import GCSDevice, pitools

        self.pitools = pitools

        ''' Setting up the PI stages '''
        self.pi = self.cfg.pi_parameters

        self.controllername = self.cfg.pi_parameters['controllername']
        self.pi_stages = self.cfg.pi_parameters['stages']
        # ('M-112K033','L-406.40DG10','M-112K033','M-116.DG','M-406.4PD','NOSTAGE')
        self.refmode = self.cfg.pi_parameters['refmode']
        # self.serialnum = ('118015439')  # Wyss Geneva
        self.serialnum = self.cfg.pi_parameters['serialnum']  # UZH Irchel H45

        self.pidevice = GCSDevice(self.controllername)
        self.pidevice.ConnectUSB(serialnum=self.serialnum)

        ''' PI startup '''

        ''' with refmode enabled: pretty dangerous
        pitools.startup(self.pidevice, stages=self.pi_stages, refmode=self.refmode)
        '''
        pitools.startup(self.pidevice, stages=self.pi_stages)

        ''' Setting PI velocities '''
        self.pidevice.VEL(self.cfg.pi_parameters['velocity'])

        self.pidevice.FRF(1)
        print('M-061 Emergency referencing hack: Waiting for referencing move')
        logger.info('M-061 Emergency referencing hack: Waiting for referencing move')
        self.block_till_controller_is_ready()
        print('M-061 Emergency referencing hack done')
        logger.info('M-061 Emergency referencing hack done')

    def __del__(self):
        try:
            '''Close the Galil connection'''
            self.xyz_stage.close()
            self.f_stage.close_stage()
        except:
            pass

    def report_position(self):
        positions = self.pidevice.qPOS(self.pidevice.axes)

        '''
        Ugly workaround to deal with non-responding stage 
        position reports: Do not update positions in 
        exceptional circumstances. 
        '''
        try:
            self.x_pos, self.y_pos, self.z_pos = self.xyz_stage.read_position()
            _, _, self.f_pos = self.f_stage.read_position()
        except:
            logger.info('Error while unpacking Galil stage position values')

            self.create_position_dict()

        self.theta_pos = positions['1']

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)
        # print(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' Galil move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        xyz_motion_dict = {}

        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                xyz_motion_dict.update({1: int(x_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                xyz_motion_dict.update({2: int(y_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                xyz_motion_dict.update({3: int(z_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if xyz_motion_dict != {}:
            self.xyz_stage.move_relative(xyz_motion_dict)

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                self.pidevice.MVR({1: theta_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                self.f_stage.move_relative({3: int(f_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if wait_until_done == True:
            self.f_stage.wait_until_done('Z')
            self.xyz_stage.wait_until_done('XYZ')
            self.pitools.waitontarget(self.pidevice)

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Galil move absolute method

        Lots of implementation details in here, should be replaced by a facade

        '''
        xyz_motion_dict = {}

        if 'x_abs' or 'y_abs' or 'z_abs' in dict:
            if 'x_abs' in dict:
                x_abs = dict['x_abs']
                x_abs = x_abs - self.int_x_pos_offset
                xyz_motion_dict.update({1: x_abs})

            if 'y_abs' in dict:
                y_abs = dict['y_abs']
                y_abs = y_abs - self.int_y_pos_offset
                xyz_motion_dict.update({2: y_abs})

            if 'z_abs' in dict:
                z_abs = dict['z_abs']
                z_abs = z_abs - self.int_z_pos_offset
                xyz_motion_dict.update({3: z_abs})

        if xyz_motion_dict != {}:
            self.xyz_stage.move_absolute(xyz_motion_dict)

        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs and self.f_max > f_abs:
                ''' Conversion to mm and command emission'''
                self.f_stage.move_absolute({3: int(f_abs)})
            else:
                self.sig_status_message.emit('Absolute movement stopped: F Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs and self.theta_max > theta_abs:
                ''' No Conversion to mm !!!! and command emission'''
                self.pidevice.MOV({1: theta_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done == True:
            self.pitools.waitontarget(self.pidevice)

    def stop(self):
        self.f_stage.stop(restart_programs=True)
        self.xyz_stage.stop(restart_programs=True)
        self.pidevice.STP(noraise=True)

    def load_sample(self):
        self.move_absolute({'y_abs': self.cfg.stage_parameters['y_load_position']})

    def unload_sample(self):
        self.move_absolute({'y_abs': self.cfg.stage_parameters['y_unload_position']})

    def go_to_rotation_position(self, wait_until_done=False):
        self.move_absolute({'x_abs': self.x_rot_position, 'y_abs': self.y_rot_position, 'z_abs': self.z_rot_position})
        if wait_until_done == True:
            self.xyz_stage.wait_until_done('XYZ')

    def block_till_controller_is_ready(self):
        '''
        Blocks further execution (especially during referencing moves)
        till the PI controller returns ready
        '''
        blockflag = True
        while blockflag:
            if self.pidevice.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)

    def execute_program(self):
        '''Executes program stored on the Galil controller'''
        self.f_stage.execute_program()
        self.xyz_stage.execute_program()


class mesoSPIM_PI_rotz_and_Galil_xyf_Stages(mesoSPIM_Stage):
    '''
    Deprecated?
    Expects following microscope configuration:

    Sample XYF movement: Galil controller with 3 axes
    Z-Movement and Rotation: PI C-884 mercury controller
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        # self.state = mesoSPIM_StateSingleton()

        self.pos_timer = QtCore.QTimer(self)
        self.pos_timer.timeout.connect(self.report_position)
        self.pos_timer.start(50)
        '''
        Galil-specific code
        '''
        from .devices.stages.galil.galilcontrol import StageControlGalil

        self.x_encodercounts_per_um = self.cfg.xyf_galil_parameters['x_encodercounts_per_um']
        self.y_encodercounts_per_um = self.cfg.xyf_galil_parameters['y_encodercounts_per_um']
        self.f_encodercounts_per_um = self.cfg.xyf_galil_parameters['f_encodercounts_per_um']

        ''' Setting up the Galil stages: XYZ '''
        self.xyf_stage = StageControlGalil(self.cfg.xyf_galil_parameters['port'], [self.x_encodercounts_per_um,
                                                                                   self.y_encodercounts_per_um,
                                                                                   self.f_encodercounts_per_um])

        ''' PI-specific code '''
        from pipython import GCSDevice, pitools

        self.pitools = pitools

        ''' Setting up the PI stages '''
        self.pi = self.cfg.pi_parameters

        self.controllername = self.cfg.pi_parameters['controllername']
        self.pi_stages = list(self.cfg.pi_parameters['stages'])
        # ('M-112K033','L-406.40DG10','M-112K033','M-116.DG','M-406.4PD','NOSTAGE')
        self.refmode = self.cfg.pi_parameters['refmode']
        # self.serialnum = ('118015439')  # Wyss Geneva
        self.serialnum = self.cfg.pi_parameters['serialnum']  # UZH Irchel H45

        self.pidevice = GCSDevice(self.controllername)
        self.pidevice.ConnectUSB(serialnum=self.serialnum)

        ''' PI startup '''

        ''' with refmode enabled: pretty dangerous
        pitools.startup(self.pidevice, stages=self.pi_stages, refmode=self.refmode)
        '''
        pitools.startup(self.pidevice, stages=self.pi_stages)

        ''' Setting PI velocities '''
        self.pidevice.VEL(self.cfg.pi_parameters['velocity'])

        ''' Reference movements '''
        print('M-406 Emergency referencing hack: Waiting for referencing move')
        logger.info('M-406 Emergency referencing hack: Waiting for referencing move')
        self.pidevice.FRF(2)
        print('M-406 Emergency referencing hack done')
        logger.info('M-406 Emergency referencing hack done')

    def __del__(self):
        try:
            '''Close the Galil connection'''
            self.xyf_stage.close()
        except:
            pass

    def report_position(self):
        positions = self.pidevice.qPOS(self.pidevice.axes)

        '''
        Ugly workaround to deal with non-responding stage 
        position reports: Do not update positions in 
        exceptional circumstances. 
        '''
        try:
            self.x_pos, self.y_pos, self.f_pos = self.xyf_stage.read_position()
        except:
            logger.info('Error while unpacking Galil stage position values')

        self.create_position_dict()

        self.z_pos = round(positions['2'] * 1000, 2)
        self.theta_pos = positions['1']

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)
        # print(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' Galil move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        xyf_motion_dict = {}

        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                xyf_motion_dict.update({1: int(x_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                xyf_motion_dict.update({2: int(y_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                z_rel = z_rel / 1000
                self.pidevice.MVR({2: z_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                self.pidevice.MVR({1: theta_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                xyf_motion_dict.update({3: int(f_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if xyf_motion_dict != {}:
            self.xyf_stage.move_relative(xyf_motion_dict)

        if wait_until_done == True:
            self.xyf_stage.wait_until_done('XYZ')
            self.pitools.waitontarget(self.pidevice)

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Galil move absolute method

        Lots of implementation details in here, should be replaced by a facade

        '''
        xyf_motion_dict = {}

        if 'x_abs' or 'y_abs' or 'f_abs' in dict:
            if 'x_abs' in dict:
                x_abs = dict['x_abs']
                x_abs = x_abs - self.int_x_pos_offset
                xyf_motion_dict.update({1: x_abs})

            if 'y_abs' in dict:
                y_abs = dict['y_abs']
                y_abs = y_abs - self.int_y_pos_offset
                xyf_motion_dict.update({2: y_abs})

            if 'f_abs' in dict:
                f_abs = dict['f_abs']
                f_abs = f_abs - self.int_f_pos_offset
                xyf_motion_dict.update({3: f_abs})

        if xyf_motion_dict != {}:
            self.xyf_stage.move_absolute(xyf_motion_dict)

        if wait_until_done == True:
            self.xyf_stage.wait_until_done('XYZ')

        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs and self.z_max > z_abs:
                ''' Conversion to mm and command emission'''
                z_abs = z_abs / 1000
                self.pidevice.MOV({2: z_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Z Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs and self.theta_max > theta_abs:
                ''' No Conversion to mm !!!! and command emission'''
                self.pidevice.MOV({1: theta_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done == True:
            self.xyf_stage.wait_until_done('XYZ')
            self.pitools.waitontarget(self.pidevice)

    def stop(self):
        self.xyf_stage.stop(restart_programs=True)
        self.pidevice.STP(noraise=True)

    def load_sample(self):
        self.xyf_stage.move_absolute({2: self.cfg.stage_parameters['y_load_position']})

    def unload_sample(self):
        self.xyf_stage.move_absolute({2: self.cfg.stage_parameters['y_unload_position']})

    def go_to_rotation_position(self, wait_until_done=False):
        ''' This has to be done in absolute coordinates of the stages to avoid problems with the 
        internal position offset (when the stage is zeroed). '''
        xy_motion_dict = {1: self.x_rot_position, 2: self.y_rot_position}
        self.xyf_stage.move_absolute(xy_motion_dict)
        self.pidevice.MOV({2: self.z_rot_position / 1000})

        if wait_until_done == True:
            self.xyf_stage.wait_until_done('XYZ')
            self.pitools.waitontarget(self.pidevice)

    def block_till_controller_is_ready(self):
        '''
        Blocks further execution (especially during referencing moves)
        till the PI controller returns ready
        '''
        blockflag = True
        while blockflag:
            if self.pidevice.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)

    def execute_program(self):
        '''Executes program stored on the Galil controller'''
        self.xyf_stage.execute_program()


class mesoSPIM_PI_rotzf_and_Galil_xy_Stages(mesoSPIM_Stage):
    '''
    Deprecated?
    Expects following microscope configuration:
    
    Sample XY movement: Galil controller with 2 axes 
    Z-Movement, F-Movement and Rotation: PI C-884 mercury controller
    '''

    def __init__(self, parent=None):
        super().__init__(parent)

        self.pos_timer = QtCore.QTimer(self)
        self.pos_timer.timeout.connect(self.report_position)
        self.pos_timer.start(50)
        '''
        Galil-specific code
        '''
        from .devices.stages.galil.galilcontrol import StageControlGalil

        self.x_encodercounts_per_um = self.cfg.xy_galil_parameters['x_encodercounts_per_um']
        self.y_encodercounts_per_um = self.cfg.xy_galil_parameters['y_encodercounts_per_um']

        ''' Setting up the Galil stages: XYZ '''
        self.xy_stage = StageControlGalil(self.cfg.xy_galil_parameters['port'], [self.x_encodercounts_per_um,
                                                                                 self.y_encodercounts_per_um])

        ''' PI-specific code '''
        from pipython import GCSDevice, pitools

        self.pitools = pitools

        ''' Setting up the PI stages '''
        self.pi = self.cfg.pi_parameters

        self.controllername = self.cfg.pi_parameters['controllername']
        self.pi_stages = list(self.cfg.pi_parameters['stages'])
        # ('M-112K033','L-406.40DG10','M-112K033','M-116.DG','M-406.4PD','NOSTAGE')
        self.refmode = self.cfg.pi_parameters['refmode']
        self.serialnum = self.cfg.pi_parameters['serialnum']

        self.pidevice = GCSDevice(self.controllername)
        self.pidevice.ConnectUSB(serialnum=self.serialnum)

        ''' PI startup '''

        ''' with refmode enabled: pretty dangerous
        pitools.startup(self.pidevice, stages=self.pi_stages, refmode=self.refmode)
        '''
        pitools.startup(self.pidevice, stages=self.pi_stages)

        ''' Setting PI velocities '''
        self.pidevice.VEL(self.cfg.pi_parameters['velocity'])

        print('M-406 Emergency referencing hack: Waiting for referencing move')
        logger.info('M-406 Emergency referencing hack: Waiting for referencing move')
        self.pidevice.FRF(2)
        print('M-406 Emergency referencing hack done')
        logger.info('M-406 Emergency referencing hack done')

        print('M-605.2DD Emergency referencing hack: Waiting for referencing move')
        logger.info('M-605.2DD  Emergency referencing hack: Waiting for referencing move')
        self.pidevice.FRF(3)
        print('M-605.2DD Emergency referencing hack done')
        logger.info('M-605.2DD Emergency referencing hack done')

        self.block_till_controller_is_ready()

    def __del__(self):
        try:
            '''Close the Galil connection'''
            self.xy_stage.close()
        except:
            pass

    def report_position(self):
        positions = self.pidevice.qPOS(self.pidevice.axes)

        '''
        Ugly workaround to deal with non-responding stage 
        position reports: Do not update positions in 
        exceptional circumstances. 
        '''
        try:
            self.x_pos, self.y_pos = self.xy_stage.read_position()
        except:
            logger.info('Error while unpacking Galil stage position values')

        self.f_pos = round(positions['3'] * 1000, 2)
        self.z_pos = round(positions['2'] * 1000, 2)
        self.theta_pos = positions['1']

        self.create_position_dict()

        self.int_x_pos = self.x_pos + self.int_x_pos_offset
        self.int_y_pos = self.y_pos + self.int_y_pos_offset
        self.int_z_pos = self.z_pos + self.int_z_pos_offset
        self.int_f_pos = self.f_pos + self.int_f_pos_offset
        self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

        self.create_internal_position_dict()

        self.sig_position.emit(self.int_position_dict)
        # print(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' Galil move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''
        xy_motion_dict = {}

        if 'x_rel' in dict:
            x_rel = dict['x_rel']
            if self.x_min < self.x_pos + x_rel and self.x_max > self.x_pos + x_rel:
                xy_motion_dict.update({1: int(x_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                xy_motion_dict.update({2: int(y_rel)})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                z_rel = z_rel / 1000
                self.pidevice.MVR({2: z_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')

        if 'theta_rel' in dict:
            theta_rel = dict['theta_rel']
            if self.theta_min < self.theta_pos + theta_rel and self.theta_max > self.theta_pos + theta_rel:
                self.pidevice.MVR({1: theta_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                f_rel = f_rel / 1000
                self.pidevice.MVR({3: f_rel})
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if xy_motion_dict != {}:
            self.xy_stage.move_relative(xy_motion_dict)

        if wait_until_done == True:
            self.xy_stage.wait_until_done('XY')
            self.pitools.waitontarget(self.pidevice)

    def move_absolute(self, dict, wait_until_done=False):
        '''
        Galil move absolute method

        Lots of implementation details in here, should be replaced by a facade

        '''
        xy_motion_dict = {}

        if 'x_abs' or 'y_abs' in dict:
            if 'x_abs' in dict:
                x_abs = dict['x_abs']
                x_abs = x_abs - self.int_x_pos_offset
                xy_motion_dict.update({1: x_abs})

            if 'y_abs' in dict:
                y_abs = dict['y_abs']
                y_abs = y_abs - self.int_y_pos_offset
                xy_motion_dict.update({2: y_abs})

        if xy_motion_dict != {}:
            self.xy_stage.move_absolute(xy_motion_dict)

        if wait_until_done == True:
            self.xy_stage.wait_until_done('XYZ')

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs and self.f_max > f_abs:
                ''' Conversion to mm and command emission'''
                f_abs = f_abs / 1000
                self.pidevice.MOV({3: f_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: F Motion limit would be reached!')

        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs and self.z_max > z_abs:
                ''' Conversion to mm and command emission'''
                z_abs = z_abs / 1000
                self.pidevice.MOV({2: z_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Z Motion limit would be reached!')

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs and self.theta_max > theta_abs:
                ''' No Conversion to mm !!!! and command emission'''
                self.pidevice.MOV({1: theta_abs})
            else:
                self.sig_status_message.emit('Absolute movement stopped: Theta Motion limit would be reached!')

        if wait_until_done == True:
            self.xy_stage.wait_until_done('XY')
            self.pitools.waitontarget(self.pidevice)

    def stop(self):
        self.xy_stage.stop(restart_programs=True)
        self.pidevice.STP(noraise=True)

    def load_sample(self):
        self.xy_stage.move_absolute({2: self.cfg.stage_parameters['y_load_position']})

    def unload_sample(self):
        self.xy_stage.move_absolute({2: self.cfg.stage_parameters['y_unload_position']})

    def go_to_rotation_position(self, wait_until_done=False):
        ''' This has to be done in absolute coordinates of the stages to avoid problems with the 
        internal position offset (when the stage is zeroed). '''
        xy_motion_dict = {1: self.x_rot_position, 2: self.y_rot_position}
        self.xy_stage.move_absolute(xy_motion_dict)
        self.pidevice.MOV({2: self.z_rot_position / 1000})

        if wait_until_done == True:
            self.xy_stage.wait_until_done('XY')
            self.pitools.waitontarget(self.pidevice)

    def block_till_controller_is_ready(self):
        '''
        Blocks further execution (especially during referencing moves)
        till the PI controller returns ready
        '''
        blockflag = True
        while blockflag:
            if self.pidevice.IsControllerReady():
                blockflag = False
            else:
                time.sleep(0.1)

    def execute_program(self):
        '''Executes program stored on the Galil controller'''
        self.xy_stage.execute_program()


class mesoSPIM_ASI_Tiger_Stage(mesoSPIM_Stage):
    '''
    It is expected that the parent class has the following signals:
        sig_move_relative = pyqtSignal(dict)
        sig_move_relative_and_wait_until_done = pyqtSignal(dict)
        sig_move_absolute = pyqtSignal(dict)
        sig_move_absolute_and_wait_until_done = pyqtSignal(dict)
        sig_zero = pyqtSignal(list)
        sig_unzero = pyqtSignal(list)
        sig_stop_movement = pyqtSignal()
        sig_mark_rotation_position = pyqtSignal()

    Also contains a QTimer that regularily sends position updates, e.g
    during the execution of movements.
    '''
    sig_pause = QtCore.pyqtSignal(bool)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.state = mesoSPIM_StateSingleton()
        from .devices.stages.asi.asicontrol import StageControlASITiger
        
        ''' Setting up the ASI stages '''
        self.asi_parameters = self.cfg.asi_parameters
        self.mesoSPIM2ASIdict = self.asi_parameters['stage_assignment'] # converts mesoSPIM stage to ASI stage designation
        # self.ASI2mesoSPIMdict = {self.mesoSPIM2ASIdict[item] : item for item in self.mesoSPIM2ASIdict} # converts ASI stage designation to mesoSPIM

        self.ttl_cards = self.asi_parameters['ttl_cards']
        self.asi_stages = StageControlASITiger(self.asi_parameters)
        self.asi_stages.sig_pause.connect(self.pause)

        assert hasattr(self.cfg, 'asi_parameters'), "Config file with stage 'TigerASI' must have 'asi_parameters' dict."
        self.ttl_motion_enabled_during_acq = self.cfg.asi_parameters['ttl_motion_enabled']
        self.ttl_motion_currently_enabled = False

        self.pos_timer.setInterval(250)
        logger.info('ASI stages initialized')
        
    def __del__(self):
        try:
            self.asi_stages.close()
        except:
            pass

    @QtCore.pyqtSlot(bool)
    def pause(self, boolean):
        state = self.state['state']
        if state == 'run_selected_acquisition' or state == 'run_acquisition_list':
            self.sig_pause.emit(boolean)

    @QtCore.pyqtSlot(dict)
    def log_slice(self, dictionary):
        slice = dictionary['current_image_in_acq']
        self.asi_stages.current_z_slice = slice

    def report_position(self):
        position_dict = self.asi_stages.read_position()
        if position_dict is not None:
            self.x_pos = position_dict[self.mesoSPIM2ASIdict['x']]
            self.y_pos = position_dict[self.mesoSPIM2ASIdict['y']]
            self.z_pos = position_dict[self.mesoSPIM2ASIdict['z']]
            self.f_pos = position_dict[self.mesoSPIM2ASIdict['f']]
            self.theta_pos = position_dict[self.mesoSPIM2ASIdict['theta']]/100

            self.create_position_dict()

            self.int_x_pos = self.x_pos + self.int_x_pos_offset
            self.int_y_pos = self.y_pos + self.int_y_pos_offset
            self.int_z_pos = self.z_pos + self.int_z_pos_offset
            self.int_f_pos = self.f_pos + self.int_f_pos_offset
            self.int_theta_pos = self.theta_pos + self.int_theta_pos_offset

            self.create_internal_position_dict()

            self.sig_position.emit(self.int_position_dict)

    def move_relative(self, dict, wait_until_done=False):
        ''' ASI move relative method
        Lots of implementation details in here, should be replaced by a facade
        '''
        motion_dict = {}
        if not self.ttl_motion_currently_enabled:
            if 'x_rel' in dict:
                x_rel = dict['x_rel']
                if self.x_min < self.x_pos + x_rel < self.x_max:
                    motion_dict.update({self.mesoSPIM2ASIdict['x'] : round(x_rel, 1)})
                else:
                    self.sig_status_message.emit('Relative movement stopped: X Motion limit would be reached!')

            if 'y_rel' in dict:
                y_rel = dict['y_rel']
                if self.y_min < self.y_pos + y_rel < self.y_max:
                    motion_dict.update({self.mesoSPIM2ASIdict['y'] : round(y_rel, 1)})
                else:
                    self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

            if 'z_rel' in dict:
                z_rel = dict['z_rel']
                if self.z_min < self.z_pos + z_rel < self.z_max:
                    motion_dict.update({self.mesoSPIM2ASIdict['z'] : round(z_rel, 1)})
                else:
                    self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')
            
            if 'theta_rel' in dict:
                theta_rel = dict['theta_rel']
                if self.theta_min < self.theta_pos + theta_rel < self.theta_max:
                    ''' 1° equals 1000 cts, but there is a factor 10 in asicontrol.py '''
                    motion_dict.update({self.mesoSPIM2ASIdict['theta'] : int(theta_rel*100)})
                else:
                    self.sig_status_message.emit('Relative movement stopped: theta Motion limit would be reached!')

            if 'f_rel' in dict:
                f_rel = dict['f_rel']
                if self.f_min < self.f_pos + f_rel < self.f_max:
                    motion_dict.update({self.mesoSPIM2ASIdict['f'] : round(f_rel, 1)})
                else:
                    self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

            if motion_dict != {}:
                self.asi_stages.move_relative(motion_dict)

            if wait_until_done:
                self.asi_stages.wait_until_done()
    
    def move_absolute(self, dict, wait_until_done=False):
        '''
        ASI move absolute method

        Lots of implementation details in here, should be replaced by a facade
        '''

        #self.adapt_position_polling_interval_to_state()
        motion_dict = {}
        if 'x_abs' in dict:
            x_abs = dict['x_abs']
            x_abs = x_abs - self.int_x_pos_offset
            if self.x_min < x_abs < self.x_max:
                motion_dict.update({self.mesoSPIM2ASIdict['x'] : round(x_abs, 1)})
            else:
                logger.error(f"The x-move is outside of min-max range, check your config file, 'x_min' and 'x_max'.")

        if 'y_abs' in dict:
            y_abs = dict['y_abs']
            y_abs = y_abs - self.int_y_pos_offset
            if self.y_min < y_abs < self.y_max:
                motion_dict.update({self.mesoSPIM2ASIdict['y'] : round(y_abs, 1)})
            else:
                logger.error(f"The y-move is outside of min-max range, check your config file, 'y_min' and 'y_max'.")
                    
        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs < self.z_max:
                motion_dict.update({self.mesoSPIM2ASIdict['z'] : round(z_abs, 1)})
            else:
                logger.error(f"The z-move is outside of min-max range, check your config file, 'z_min' and 'z_max'.")

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs < self.f_max:
                motion_dict.update({self.mesoSPIM2ASIdict['f'] : round(f_abs, 1)})
            else:
                logger.error(f"The f-move is outside of min-max range, check your config file, 'f_min' and 'f_max'.")

        if 'theta_abs' in dict:
            theta_abs = dict['theta_abs']
            theta_abs = theta_abs - self.int_theta_pos_offset
            if self.theta_min < theta_abs < self.theta_max:
                ''' 1° equals 1000 cts, but there is a factor 10 in asicontrol.py '''
                motion_dict.update({self.mesoSPIM2ASIdict['theta'] : int(theta_abs*100)})
            else:
                logger.error(f"The theta-move is outside of min-max range, check your config file, 'theta_min' and 'theta_max'.")

        if motion_dict:
            self.asi_stages.move_absolute(motion_dict)
        
        if wait_until_done is True:
            self.asi_stages.wait_until_done()
        
    def stop(self):
        self.asi_stages.stop()

    def load_sample(self):
        y_abs = self.cfg.stage_parameters['y_load_position']
        self.move_absolute({'y_abs':round(y_abs)})

    def unload_sample(self):
        y_abs = self.cfg.stage_parameters['y_unload_position']
        self.move_absolute({'y_abs':round(y_abs)})

    def go_to_rotation_position(self, wait_until_done=False):
        x_abs = self.x_rot_position
        y_abs = self.y_rot_position
        z_abs = self.z_rot_position

        self.move_absolute({'x_abs':x_abs, 'y_abs':y_abs, 'z_abs':z_abs})

        if wait_until_done == True:
            self.asi_stages.wait_until_done()

    def enable_ttl_motion(self, boolean):
        if self.ttl_motion_enabled_during_acq:
            self.asi_stages.enable_ttl_mode(self.ttl_cards, boolean)
            self.ttl_motion_currently_enabled = boolean
            logger.info('TTL Motion currently enabled: '+str(boolean))
            self.state['ttl_movement_enabled_during_acq'] = boolean
        

class mesoSPIM_ASI_MS2000_Stage(mesoSPIM_Stage):
    '''

    It is expected that the parent class has the following signals:
        sig_move_relative = pyqtSignal(dict)
        sig_move_relative_and_wait_until_done = pyqtSignal(dict)
        sig_move_absolute = pyqtSignal(dict)
        sig_move_absolute_and_wait_until_done = pyqtSignal(dict)
        sig_zero = pyqtSignal(list)
        sig_unzero = pyqtSignal(list)
        sig_stop_movement = pyqtSignal()
        sig_mark_rotation_position = pyqtSignal()

    Also contains a QTimer that regularily sends position updates, e.g
    during the execution of movements.

    This implements an ASI MS2000 controller for a setup with the following configuration
    * ASI X Stage is equivalent to the mesoSPIM z-stage (moved during stacks direction)
    * ASI Y Stage is equivalent to the mesoSPIM y-stage  
    * ASI Z-Stage is equivalent to the mesoSPIM f-stage (focus)

    '''
    sig_pause = QtCore.pyqtSignal(bool)

    def __init__(self, parent = None):
        super().__init__(parent)

        self.state = mesoSPIM_StateSingleton()
        '''
        ASI-specific code
        '''
        from devices.stages.asi.asicontrol import StageControlASITiger
        
        ''' Setting up the ASI stages '''
        self.asi_parameters = self.cfg.asi_parameters
        self.mesoSPIM2ASIdict = self.asi_parameters['stage_assignment'] # converts mesoSPIM stage to ASI stage designation
        # self.ASI2mesoSPIMdict = {self.mesoSPIM2ASIdict[item] : item for item in self.mesoSPIM2ASIdict} # converts ASI stage designation to mesoSPIM

        self.asi_stages = StageControlASITiger(self.asi_parameters)
        self.asi_stages.sig_pause.connect(self.pause)
        self.pos_timer.setInterval(100)

        logger.info('mesoSPIM_Stages: ASI stages initialized')

    def __del__(self):
        try:
            '''Close the ASI connection'''
            self.asi_stages.close()
        except:
            pass

    @QtCore.pyqtSlot(bool)
    def pause(self,boolean):
        state = self.state['state']
        if state == 'run_selected_acquisition' or state == 'run_acquisition_list':
            self.sig_pause.emit(boolean)

    @QtCore.pyqtSlot(dict)
    def log_slice(self, dictionary):
        slice = dictionary['current_image_in_acq']
        self.asi_stages.current_z_slice = slice

    def report_position(self):
        position_dict = self.asi_stages.read_position()
        if position_dict is not None:
            self.y_pos = position_dict[self.mesoSPIM2ASIdict['y']]
            self.z_pos = position_dict[self.mesoSPIM2ASIdict['z']]
            self.f_pos = position_dict[self.mesoSPIM2ASIdict['f']]
            
            self.create_position_dict()

            self.int_y_pos = self.y_pos + self.int_y_pos_offset
            self.int_z_pos = self.z_pos + self.int_z_pos_offset
            self.int_f_pos = self.f_pos + self.int_f_pos_offset

            self.create_internal_position_dict()

            self.sig_position.emit(self.int_position_dict)


    def move_relative(self, dict, wait_until_done=False):
        ''' ASI move relative method

        Lots of implementation details in here, should be replaced by a facade
        '''

        '''
        Report position 

        '''
        #self.adapt_position_polling_interval_to_state()

        motion_dict = {}

        if 'y_rel' in dict:
            y_rel = dict['y_rel']
            if self.y_min < self.y_pos + y_rel and self.y_max > self.y_pos + y_rel:
                motion_dict.update({self.mesoSPIM2ASIdict['y'] : round(y_rel, 1)})
            else:
                self.sig_status_message.emit('Relative movement stopped: Y Motion limit would be reached!')

        if 'z_rel' in dict:
            z_rel = dict['z_rel']
            if self.z_min < self.z_pos + z_rel and self.z_max > self.z_pos + z_rel:
                motion_dict.update({self.mesoSPIM2ASIdict['z'] : round(z_rel, 1)})
            else:
                self.sig_status_message.emit('Relative movement stopped: z Motion limit would be reached!')
        
        if 'f_rel' in dict:
            f_rel = dict['f_rel']
            if self.f_min < self.f_pos + f_rel and self.f_max > self.f_pos + f_rel:
                motion_dict.update({self.mesoSPIM2ASIdict['f'] : round(f_rel, 1)})
            else:
                self.sig_status_message.emit('Relative movement stopped: f Motion limit would be reached!')

        if motion_dict != {}:
            self.asi_stages.move_relative(motion_dict)

        if wait_until_done is True:
            self.asi_stages.wait_until_done()
    
    def move_absolute(self, dict, wait_until_done=False):
        '''
        ASI move absolute method

        Lots of implementation details in here, should be replaced by a facade
        '''
        #self.adapt_position_polling_interval_to_state()

        motion_dict = {}

        if 'y_abs' in dict:
            y_abs = dict['y_abs']
            y_abs = y_abs - self.int_y_pos_offset
            if self.y_min < y_abs and self.y_max > y_abs:
                motion_dict.update({self.mesoSPIM2ASIdict['y'] : round(y_abs, 1)})
                    
        if 'z_abs' in dict:
            z_abs = dict['z_abs']
            z_abs = z_abs - self.int_z_pos_offset
            if self.z_min < z_abs and self.z_max > z_abs:
                motion_dict.update({self.mesoSPIM2ASIdict['z'] : round(z_abs, 1)})

        if 'f_abs' in dict:
            f_abs = dict['f_abs']
            f_abs = f_abs - self.int_f_pos_offset
            if self.f_min < f_abs and self.f_max > f_abs:
                motion_dict.update({self.mesoSPIM2ASIdict['f'] : round(f_abs, 1)})

        if motion_dict != {}:
            self.asi_stages.move_absolute(motion_dict)
        
        if wait_until_done == True:
            self.asi_stages.wait_until_done()
        
    def stop(self):
        self.asi_stages.stop()

    def load_sample(self):
        message = 'ASI MS-2000 Stage: Sample loading not implemented!'
        print(message)
        logger.info(message)

    def unload_sample(self):
        message = 'ASI MS-2000 Stage: Sample unloading not implemented!'
        print(message)
        logger.info(message)

    def go_to_rotation_position(self, wait_until_done=False):
        message = 'ASI MS-2000 Stage: Rotation position not implemented!'
        print(message)
        logger.info(message)


"""
mesoSPIM Module for enabling single laser lines via NI-DAQmx

Authors: Fabian Voigt, Nikita Vladimirov
"""

import nidaqmx
from nidaqmx.constants import LineGrouping

class mesoSPIM_LaserEnabler:
    ''' Class for interacting with the laser enable DO lines via NI-DAQmx
    This uses the property of NI-DAQmx-outputs to keep their last digital state or
    analog voltage for as long the device is not powered down. This means that
    the NI tasks are closed after calls to "enable", "disable", etc which in turn
    means that this class is not intended for fast switching in complicated waveforms.
    Needs a dictionary which combines laser wavelengths and device outputs
    in the form:
    {'488 nm': 'PXI1Slot4/port0/line2',
    '515 nm': 'PXI1Slot4/port0/line3'}
    '''
    def __init__(self, laserdict):
        self.laserenablestate = 'None'
        self.laserdict = laserdict

        # get a value in the laserdict to get the general device string
        self.laserenable_device = ''
        self.laser_keys_sorted = sorted(laserdict.keys())
        for key in self.laser_keys_sorted:
            self.laserenable_device += laserdict[key] + ','
        self.disable_all()         # Make sure that all the Lasers are off upon initialization:

    def _check_if_laser_in_laserdict(self, laser):
        '''Checks if the laser designation (string) given as argument exists in the laserdict'''
        if laser in self.laserdict:
            return True
        else:
            raise ValueError('Laser not in the configuration')

    def enable(self, laser):
        '''Enables a single laser line. All other lines are switched off.'''
        if self._check_if_laser_in_laserdict(laser):
            command_list = [False]*len(self.laserdict)
            ind_line_on = list(self.laser_keys_sorted).index(laser)
            command_list[ind_line_on] = True
            with nidaqmx.Task() as task:
                task.do_channels.add_do_chan(self.laserenable_device, line_grouping=LineGrouping.CHAN_PER_LINE)
                task.write(command_list, auto_start=True)
            self.laserenablestate = laser
        else:
            pass

    def disable_all(self):
        '''Disables all laser lines.'''
        with nidaqmx.Task() as task:
            task.do_channels.add_do_chan(self.laserenable_device, line_grouping=LineGrouping.CHAN_PER_LINE)
            task.write([False]*len(self.laserdict), auto_start=True)
        self.laserenablestate = 'off'

    def state(self):
        """ Returns laserline if a laser is on, otherwise "False" """
        return self.laserenablestate



import numpy as np

'''
This is a hacked confige file, where physical shutter is used to control 568 laser port, and in software it is controlled by laser-enable line of 568 laser.
Valid only for a demo version of denchtop mesoSPIM, to recycle axisting 568nm laser.
'''

ui_options = {'dark_mode' : True, # Dark mode: Renders the UI dark if enabled
              'enable_x_buttons' : True, # Here, specific sets of UI buttons can be disabled
              'enable_y_buttons' : True,
              'enable_z_buttons' : True,
              'enable_f_buttons' : True,
              'enable_rotation_buttons' : True,
              'enable_loading_buttons' : True,
              'window_pos': (100, 100), # position of the main window on the screen, top left corner.
               }

'''
Waveform output for Galvos, ETLs etc.
'''

waveformgeneration = 'NI' # 'DemoWaveFormGeneration' or 'NI'

'''
Card designations need to be the same as in NI MAX, if necessary, use NI MAX
to rename your cards correctly.

A standard mesoSPIM configuration uses two cards:

PXI1Slot5 (old PXI_6229_2) is responsible for the lasers (analog intensity control)
PXI1Slot2 (old PXI_6229_1) is responsible for the shutters, ETL waveforms and galvo waveforms
'''

acquisition_hardware = {'master_trigger_out_line' : 'PXI1Slot2/port0/line1',
                        'camera_trigger_source' : '/PXI1Slot2/PFI0',
                        'camera_trigger_out_line' : '/PXI1Slot2/ctr0',
                        'stage_trigger_source' : '/PXI1Slot2/PFI0',
                        'stage_trigger_out_line' : '/PXI1Slot5/ctr0',
                        'galvo_etl_task_line' : 'PXI1Slot2/ao0:3',
                        'galvo_etl_task_trigger_source' : '/PXI1Slot2/PFI0',
                        'laser_task_line' :  'PXI1Slot5/ao0:3',
                        'laser_task_trigger_source' : '/PXI1Slot2/PFI0'}

'''
Human interface device (Joystick)
'''
sidepanel = 'Demo' #'Demo' or 'FarmSimulator'

'''
Digital laser enable lines
'''

laser = 'NI' # 'Demo' or 'NI'

''' Laser blanking indicates whether the laser enable lines should be set to LOW between
individual images or stacks. This is helpful to avoid laser bleedthrough between images caused by insufficient
modulation depth of the analog input (even at 0V, some laser light is still emitted).
'''
laser_blanking = 'images' # if 'images', laser is off before and after every image; if 'stacks', before and after stacks.

''' The laserdict keys are the laser designation that will be shown
in the user interface '''

laserdict = {'488 nm': 'PXI1Slot5/port0/line2',
             '520 nm': 'PXI1Slot5/port0/line3',
             '568 nm': 'PXI1Slot5/port0/line5', # do not use second BNC block, P0.1 <-> User1 assignment fails in hardware
             '638 nm': 'PXI1Slot5/port0/line4',
             }

'''
Assignment of the analog outputs of the Laser card to the channels
The Empty slots are placeholders.
'''

laser_designation = {'488 nm' : 0,
                     '520 nm' : 1,
                     '568 nm' : 3, # empty terminal, because this 568 is not modulated
                     '638 nm' : 2,
                      }

'''
Assignment of the galvos and ETLs to the 6229 AO channels.
'''

galvo_etl_designation = {'Galvo-L' : 0,
                         'Galvo-R' : 1,
                         'ETL-L' : 2,
                         'ETL-R' : 3,
                         }

'''
Shutter configuration
'''

shutter = 'NI' # 'Demo' or 'NI'
shutterdict = {'shutter_left' : 'PXI1Slot2/port0/line0', # empty terminal, because shutter is used for 568 control
              'shutter_right' : 'PXI1Slot2/port2/line0', # flip mirror control, labeled PFI8 on the BNC connector
              }

''' A bit of a hack: Shutteroptions for the GUI '''
shutteroptions = ('Left','Right')

''' A bit of a hack: Assumes that the shutter_left line is the general shutter
and the shutter_right line is the left/right switch (Right==True)'''

shutterswitch = True # Assumes that the shutter_left line is the general shutter

'''
Camera configuration
'''

'''
For a DemoCamera, only the following options are necessary
(x_pixels and y_pixels can be chosen arbitrarily):

camera_parameters = {'x_pixels' : 1024,
                     'y_pixels' : 1024,
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4]}

For a Hamamatsu Orca Flash 4.0 V2 or V3, the following parameters are necessary:

camera_parameters = {'x_pixels' : 2048,
                     'y_pixels' : 2048,
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'camera_id' : 0,
                     'sensor_mode' : 12,    # 12 for progressive
                     'defect_correct_mode': 2,
                     'binning' : '1x1',
                     'readout_speed' : 1,
                     'trigger_active' : 1,
                     'trigger_mode' : 1, # it is unclear if this is the external lightsheeet mode - how to check this?
                     'trigger_polarity' : 2, # positive pulse
                     'trigger_source' : 2, # external
                    }

For a Photometrics Iris 15, the following parameters are necessary:

camera_parameters = {'x_pixels' : 5056,
                     'y_pixels' : 2960,
                     'x_pixel_size_in_microns' : 4.25,
                     'y_pixel_size_in_microns' : 4.25,
                     'subsampling' : [1,2,4],
                     'speed_table_index': 0,
                     'exp_mode' : 'Edge Trigger', # Lots of options in PyVCAM --> see constants.py
                     'readout_port': 0,
                     'gain_index': 1,
                     'exp_out_mode': 4, # 4: line out
                     'binning' : '1x1',
                     'scan_mode' : 1, # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
                     'scan_direction' : 0, # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}
                     'scan_line_delay' : 6, # 10.26 us x factor, a factor = 6 equals 71.82 us
                    }

For a Photometrics Prime BSI Express, the following parameters are necessary:

camera_parameters = {'x_pixels' : 2048, #5056
                     'y_pixels' : 2048, # 2960
                     'x_pixel_size_in_microns' : 6.5,
                     'y_pixel_size_in_microns' : 6.5,
                     'subsampling' : [1,2,4],
                     'speed_table_index': 1, # 1 for 100 MHz
                     'exp_mode' : 'Edge Trigger', # Lots of options in PyVCAM --> see constants.py
                     'readout_port': 0,
                     'gain_index': 1, # Enable HDR mode
                     'exp_out_mode': 4, # 4: line out
                     'binning' : '1x1',
                     'scan_mode' : 1, # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
                     'scan_direction' : 0, # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}
                     'scan_line_delay' : 3, # 11.2 us x factor, a factor = 3 equals 33.6 us
                    }

'''
camera = 'Photometrics' # 'DemoCamera' or 'HamamatsuOrcaFlash' or 'Photometrics'

camera_parameters = {'x_pixels' : 5056,
                     'y_pixels' : 2960,
                     'x_pixel_size_in_microns' : 4.25,
                     'y_pixel_size_in_microns' : 4.25,
                     'subsampling' : [1,2,4],
                     'speed_table_index': 0,
                     'exp_mode' : 'Edge Trigger', # Lots of options in PyVCAM --> see constants.py
                     'readout_port': 0,
                     'gain_index': 1,
                     'exp_out_mode': 4, # 4: line out
                     'binning' : '1x1',
                     'scan_mode' : 1, # Scan mode options: {'Auto': 0, 'Line Delay': 1, 'Scan Width': 2}
                     'scan_direction' : 0, # Scan direction options: {'Down': 0, 'Up': 1, 'Down/Up Alternate': 2}
                     'scan_line_delay' : 6, # 10.26 us x factor, a factor = 6 equals 71.82 us
                    }

binning_dict = {'1x1': (1,1), '2x2':(2,2), '4x4':(4,4)}

'''
Stage configuration
'''

'''
The stage_parameter dictionary defines the general stage configuration, initial positions,
and safety limits. The rotation position defines a XYZ position (in absolute coordinates)
where sample rotation is safe. Additional hardware dictionaries (e.g. pi_parameters)
define the stage configuration details.
'''

stage_parameters = {'stage_type' : 'TigerASI', # 'DemoStage','PI','TigerASI' or other configs found in mesoSPIM_serial.py
                    'startfocus' : -13000,
                    'y_load_position': 10000,
                    'y_unload_position': -23000,
                    'ttl_motion_enabled': True, # Controls whether stage movement is triggered by TTL pulses during acquisitions. Only available for ASI stages.
                    'x_max' : 51000,
                    'x_min' : -46000,
                    'y_max' : 160000,
                    'y_min' : -160000,
                    'z_max' : 99000,
                    'z_min' : -99000,
                    'f_max' : 99000,
                    'f_min' : -99000,
                    'theta_max' : 999,
                    'theta_min' : -999,
                    'x_rot_position': 0,
                    'y_rot_position': 0,
                    'z_rot_position': 0,
                    }
'''
Depending on the stage hardware, further dictionaries define further details of the stage configuration

For a standard mesoSPIM V4 with PI stages, the following pi_parameters are necessary (replace the
serialnumber with the one of your controller):
'''
'''
pi_parameters = {'controllername' : 'C-884',
                 'stages' : ('M-605.2DD','L-406.40DG10','M-112K033','M-116.DG','M-112K033','NOSTAGE'), # M-605.2DD, M-112K033
                 'refmode' : ('FRF',),
                 'serialnum' : ('119046748'),
                 }
'''

'''
For a standard mesoSPIM V5 with PI stages, the following pi_parameters are necessary (replace the
serialnumber with the one of your controller):

pi_parameters = {'controllername' : 'C-884',
                 'stages' : ('L-509.20DG10','L-509.40DG10','L-509.20DG10','M-060.DG','M-406.4PD','NOSTAGE'),
                 'refmode' : ('FRF',),
                 'serialnum' : ('118015799'),
'''

'''
For a benchtop mesoSPIM with an ASI Tiger controller, the following parameters are necessary.
The stage assignment dictionary assigns a mesoSPIM stage (xyzf and theta - dict key) to an ASI stage (XYZ etc)
which are the values of the dict.
'''

asi_parameters = {'COMport' : 'COM23',
                  'baudrate' : 115200,
                  # Important: The stage assignment dict has to be written in the order of installation of ASI cards
                  # This is necessary as the ASI controller always reports the stage position in the same order after a "W XYZVWT" command, no
                  # matter what order of stage designations is sent
                  # V -> (W, not necessary, because linked to V) -> Z -> T -> X -> Y
                  # X -> Y -> Z -> T -> V -> W
                  # {'y':'V', 'z':'Z', 'theta':'F', 'x':'X', 'f':'Y',} is an interim assignment - here, the Z/T card got a new firmware for
                  'stage_assignment': {'y':'V', 'z':'Z', 'theta':'R', 'x':'X', 'f':'Y'},
                  'encoder_conversion': {'V': 10., 'Z': 10., 'R': 100., 'X': 10., 'Y': 10.}, # num of encoder counts per um or degree, depending on stage type.
                  # List of card IDs that have to be triggered in TTL mode (see stage_parameters) These are usually the
                  # cards controlling the sample z-movement and the focus stage (for focus tracking acquisitions).
                  # For an MS-2000 controller, set this parameter to None
                  'ttl_cards':(2,3),
                  }

'''
Filterwheel configuration
'''

'''
For a DemoFilterWheel, no COMport needs to be specified, for a Ludl Filterwheel,
a valid COMport is necessary.
'''
filterwheel_parameters = {'filterwheel_type' : 'DemoFilterWheel', # 'DemoFilterWheel' or 'Ludl' or 'DynamixelFilterWheel'
                          'servo_id' :  1,
                          'COMport' : 'COM15',
                          'baudrate' : 115200}

# Ludl marking 10 = position 0

'''

A Ludl double filter wheel can be
'''

filterdict = {'405-488-561-640-Quadrupleblock' : 0, # Every config should contain this
              '533 540/15' : 1,
              '576 605/65' : 2,
              'Filter 4' : 3,
              'Empty-Alignment' : 4,
              }

filterpositiondict = {  0 : 1515,
                        1 : 2340,
                        2 : 3255,
                        3 : 3975,
                        4 : 4780,
                    }

'''
Zoom configuration
'''

'''
For the DemoZoom, servo_id, COMport and baudrate do not matter. For a Dynamixel zoom,
these values have to be there
'''
zoom_parameters = {'zoom_type' : 'DemoZoom', # 'DemoZoom' or 'Dynamixel'
                   'servo_id' :  1,
                   'COMport' : 'COM9',
                   'baudrate' : 115200} # 57142

'''
The keys in the zoomdict define what zoom positions are displayed in the selection box
(combobox) in the user interface.
'''

zoomdict = {'1x' : 0,
            }
'''
Pixelsize in micron
'''
pixelsize = {'1x' : 4.72,
            }

'''
Initial acquisition parameters

Used as initial values after startup

When setting up a new mesoSPIM, make sure that:
* 'max_laser_voltage' is correct (5 V for Toptica MLEs, 10 V for Omicron SOLE)
* 'galvo_l_amplitude' and 'galvo_r_amplitude' (in V) are correct (not above the max input allowed by your galvos)
* all the filepaths exist
* the initial filter exists in the filter dictionary above
'''

startup = {
'state' : 'init', # 'init', 'idle' , 'live', 'snap', 'running_script'
'samplerate' : 100000,
'sweeptime' : 0.26734,
'position' : {'x_pos':0,'y_pos':0,'z_pos':0,'f_pos':0,'theta_pos':0},
'ETL_cfg_file' : 'config/etl_parameters/ETL-parameters.csv',
'filepath' : '/tmp/file.raw',
'folder' : '/tmp/',
'snap_folder' : '/tmp/',
'file_prefix' : '',
'file_suffix' : '000001',
'zoom' : '1x',
'pixelsize' : 6.55,
'laser' : '488 nm',
'max_laser_voltage':3.3,
'intensity' : 3,
'shutterstate':False, # Is the shutter open or not?
'shutterconfig':'Left', # Can be "Left", "Right","Both","Interleaved"
'laser_interleaving':False,
'filter' : '405-488-561-640-Quadrupleblock',
'etl_l_delay_%' : 7.5,
'etl_l_ramp_rising_%' : 83,
'etl_l_ramp_falling_%' : 2.5,
'etl_l_amplitude' : 0.7,
'etl_l_offset' : 2.3,
'etl_r_delay_%' : 2.5,
'etl_r_ramp_rising_%' : 5,
'etl_r_ramp_falling_%' : 83,
'etl_r_amplitude' : 0.65,
'etl_r_offset' : 2.36,
'galvo_l_frequency' : 199.9,
'galvo_l_amplitude' : 8.0,
'galvo_l_offset' : 0,
'galvo_l_duty_cycle' : 50,
'galvo_l_phase' : np.pi/7,
'galvo_r_frequency' : 199.9,
'galvo_r_amplitude' : 8.0,
'galvo_r_offset' : 0,
'galvo_r_duty_cycle' : 50,
'galvo_r_phase' : np.pi/7,
'laser_l_delay_%' : 10,
'laser_l_pulse_%' : 87,
'laser_l_max_amplitude_%' : 100,
'laser_r_delay_%' : 10,
'laser_r_pulse_%' : 87,
'laser_r_max_amplitude_%' : 100,
'stage_trigger_delay_%' : 92.5, # Set to 92.5 for stage triggering exactly after the ETL sweep
'stage_trigger_pulse_%' : 1,
'camera_delay_%' : 10,
'camera_pulse_%' : 1,
'camera_exposure_time':0.02,
'camera_line_interval':0.000075,
'camera_display_live_subsampling': 2,
'camera_display_snap_subsampling': 1,
'camera_display_acquisition_subsampling': 2,
'camera_binning':'1x1',
'camera_sensor_mode':'ASLM',
'average_frame_rate': 3.44,
}



'''
Little helper script to execute the program stored on a Galil controller

This is useful to reinitialize the mesoSPIM Galil handcontroller after 
the program execution stopped.
'''
self.execute_galil_program()


'''
mesoSPIM Image Writer class, intended to run in the Camera Thread and handle file I/O
'''

import os
import time
import numpy as np
import tifffile
import logging
logger = logging.getLogger(__name__)
import sys
from PyQt5 import QtCore
from distutils.version import StrictVersion
from .mesoSPIM_State import mesoSPIM_StateSingleton
import npy2bdv
from .utils.acquisitions import AcquisitionList, Acquisition


class mesoSPIM_ImageWriter(QtCore.QObject):
    def __init__(self, parent=None):
        '''Image and metadata writer class. Parent is mesoSPIM_Camera() object'''
        super().__init__()

        self.parent = parent # a mesoSPIM_Camera() object
        self.cfg = parent.cfg

        self.state = mesoSPIM_StateSingleton()
        self.running = False

        self.x_pixels = self.cfg.camera_parameters['x_pixels']
        self.y_pixels = self.cfg.camera_parameters['y_pixels']

        self.binning_string = self.cfg.camera_parameters['binning'] # Should return a string in the form '2x4'
        self.x_binning = int(self.binning_string[0])
        self.y_binning = int(self.binning_string[2])

        self.x_pixels = int(self.x_pixels / self.x_binning)
        self.y_pixels = int(self.y_pixels / self.y_binning)

        self.file_extension = ''
        self.bdv_writer = self.tiff_writer = self.tiff_mip_writer = self.mip_image = None
        self.tiff_aliases = ('.tif', '.tiff')
        self.bigtiff_aliases = ('.btf', '.tf2', '.tf8')
        self.check_versions()

    def check_versions(self):
        """Take care of API changes in different library versions"""
        if StrictVersion(tifffile.__version__) < StrictVersion('2020.9.30'):
            self.tiff_write = tifffile.TiffWriter.save
            print(f"Warning: you are using outdated version of tifffile library {tifffile.__version__}. "
                  f"Upgrade to Python 3.7 and pip-install the latest tifffile version.")
        else:
            self.tiff_write = tifffile.TiffWriter.write

        tifffile.TiffWriter.write = self.tiff_write # rename the entire class method if necessary

    def prepare_acquisition(self, acq, acq_list):
        self.folder = acq['folder']
        self.filename = acq['filename']
        self.path = os.path.realpath(self.folder+'/'+self.filename)
        self.file_root, self.file_extension = os.path.splitext(self.path)
        logger.info(f'Image Writer: Save path: {self.path}')

        self.binning_string = self.state['camera_binning'] # Should return a string in the form '2x4'
        self.x_binning = int(self.binning_string[0])
        self.y_binning = int(self.binning_string[2])

        self.x_pixels = int(self.x_pixels / self.x_binning)
        self.y_pixels = int(self.y_pixels / self.y_binning)
        self.max_frame = acq.get_image_count()

        if self.file_extension == '.h5':
            if hasattr(self.cfg, "hdf5"):
                subsamp = self.cfg.hdf5['subsamp']
                compression = self.cfg.hdf5['compression']
                flip_flags = self.cfg.hdf5['flip_xyz']
            else:
                subsamp = ((1, 1, 1),)
                compression = None
                flip_flags = (False, False, False)
            # create writer object if the view is first in the list
            if acq == acq_list[0]:
                self.bdv_writer = npy2bdv.BdvWriter(self.path,
                                                    nilluminations=acq_list.get_n_shutter_configs(),
                                                    nchannels=acq_list.get_n_lasers(),
                                                    nangles=acq_list.get_n_angles(),
                                                    ntiles=acq_list.get_n_tiles(),
                                                    blockdim=((1, 256, 256),),
                                                    subsamp=subsamp,
                                                    compression=compression)
            # x and y need to be exchanged to account for the image rotation
            shape = (self.max_frame, self.x_pixels, self.y_pixels)
            px_size_um = self.cfg.pixelsize[acq['zoom']]
            sign_xyz = (1 - np.array(flip_flags)) * 2 - 1
            if hasattr(self.cfg, "hdf5") and ('transpose_xy' in self.cfg.hdf5.keys())\
                    and self.cfg.hdf5['transpose_xy'] is False:
                tile_translation = (sign_xyz[0] * acq['x_pos'] / px_size_um,
                                    sign_xyz[1] * acq['y_pos'] / px_size_um,
                                    sign_xyz[2] * acq['z_start'] / acq['z_step'])
            else: # Hamamatsu Orca Flash 4 backward compatibility
                tile_translation = (sign_xyz[1] * acq['y_pos'] / px_size_um,
                                    sign_xyz[0] * acq['x_pos'] / px_size_um,
                                    sign_xyz[2] * acq['z_start'] / acq['z_step'])
            affine_matrix = np.array(((1.0, 0.0, 0.0, tile_translation[0]),
                                      (0.0, 1.0, 0.0, tile_translation[1]),
                                      (0.0, 0.0, 1.0, tile_translation[2])))
            self.bdv_writer.append_view(stack=None, virtual_stack_dim=shape,
                                        illumination=acq_list.find_value_index(acq['shutterconfig'], 'shutterconfig'),
                                        channel=acq_list.find_value_index(acq['laser'], 'laser'),
                                        angle=acq_list.find_value_index(acq['rot'], 'rot'),
                                        tile=acq_list.get_tile_index(acq),
                                        voxel_units='um',
                                        voxel_size_xyz=(px_size_um, px_size_um, acq['z_step']),
                                        calibration=(1.0, 1.0, acq['z_step']/px_size_um),
                                        m_affine=affine_matrix,
                                        name_affine="Translation to Regular Grid"
                                        )
        elif self.file_extension == '.raw':
            self.fsize = self.x_pixels*self.y_pixels
            self.xy_stack = np.memmap(self.path, mode="write", dtype=np.uint16, shape=self.fsize * self.max_frame)

        elif self.file_extension in self.tiff_aliases:
            self.tiff_writer = tifffile.TiffWriter(self.path, imagej=True)

        elif self.file_extension in self.bigtiff_aliases:
            self.tiff_writer = tifffile.TiffWriter(self.path, bigtiff=True)

        if acq['processing'] == 'MAX' and self.file_extension in (('.raw',) + self.tiff_aliases + self.bigtiff_aliases):
            self.tiff_mip_writer = tifffile.TiffWriter(self.file_root + "_MAX.tiff", imagej=True)
            self.mip_image = np.zeros((self.y_pixels, self.x_pixels), 'uint16')

        self.cur_image = 0
        self.running = True

    def write_image(self, image, acq, acq_list):
        if self.running:
            xy_res = (1./self.cfg.pixelsize[acq['zoom']], 1./self.cfg.pixelsize[acq['zoom']])
            if self.file_extension == '.h5':
                self.bdv_writer.append_plane(plane=image, z=self.cur_image,
                                             illumination=acq_list.find_value_index(acq['shutterconfig'], 'shutterconfig'),
                                             channel=acq_list.find_value_index(acq['laser'], 'laser'),
                                             angle=acq_list.find_value_index(acq['rot'], 'rot'),
                                             tile=acq_list.get_tile_index(acq)
                                             )
            elif self.file_extension == '.raw':
                self.xy_stack[self.cur_image*self.fsize:(self.cur_image+1)*self.fsize] = image.flatten()
            elif self.file_extension in self.tiff_aliases + self.bigtiff_aliases:
                self.tiff_writer.write(image[np.newaxis,...], contiguous=True, resolution=xy_res,
                                       metadata={'spacing': acq['z_step'], 'unit': 'um'})

            if acq['processing'] == 'MAX' and self.file_extension in (('.raw',) + self.tiff_aliases + self.bigtiff_aliases):
                self.mip_image = np.maximum(self.mip_image, image)

            self.cur_image += 1
        else:
            logger.info("No image, running terminated")

    def abort_writing(self):
        """Terminate writing and close all files if STOP button is pressed"""
        if self.running:
            try:
                if self.file_extension == '.h5':
                    self.bdv_writer.close()
                elif self.file_extension == '.raw':
                    del self.xy_stack
                elif self.file_extension in (self.tiff_aliases + self.bigtiff_aliases):
                    self.tiff_writer.close()
                self.metadata_file.close()
            except Exception as e:
                logger.error(f'{e}')
            print("Writing terminated, files closed")
            self.running = False
        else:
            pass

    def end_acquisition(self, acq, acq_list):
        logger.info("end_acquisition() started")
        if self.file_extension == '.h5':
            if acq == acq_list[-1]:
                try:
                    self.bdv_writer.set_attribute_labels('channel', tuple(acq_list.get_unique_attr_list('laser')))
                    self.bdv_writer.set_attribute_labels('illumination', tuple(acq_list.get_unique_attr_list('shutterconfig')))
                    self.bdv_writer.set_attribute_labels('angle', tuple(acq_list.get_unique_attr_list('rot')))
                    self.bdv_writer.write_xml()
                except:
                    logger.error(f'HDF5 XML could not be written: {sys.exc_info()}')
                try:
                    self.bdv_writer.close()
                except:
                    logger.error(f'HDF5 file could not be closed: {sys.exc_info()}')
        elif self.file_extension == '.raw':
            try:
                del self.xy_stack
            except Exception as e:
                logger.error(f'{e}')
        elif self.file_extension in (self.tiff_aliases + self.bigtiff_aliases):
            try:
                self.tiff_writer.close()
            except Exception as e:
                logger.error(f'{e}')

        if acq['processing'] == 'MAX' and self.file_extension in (('.raw',) + self.tiff_aliases + self.bigtiff_aliases):
            try:
                self.tiff_mip_writer.write(self.mip_image)
                self.tiff_mip_writer.close()
            except Exception as e:
                logger.error(f'{e}')

        self.running = False

    def write_snap_image(self, image):
        timestr = time.strftime("%Y%m%d-%H%M%S")
        filename = timestr + '.tif'
        path = self.state['snap_folder'] + '/' + filename
        if os.path.exists(self.state['snap_folder']):
            try:
                tifffile.imsave(path, image, photometric='minisblack')
                self.write_snap_metadata(path)
            except Exception as e:
                logger.error(f"{e}")
        else:
            print(f"Error: Snap folder does not exist: {self.state['snap_folder']}. Choose it from the menu.")

    def write_line(self, file, key='', value=''):
        ''' Little helper method to write a single line with a key and value for metadata
        Adds a line break at the end.
        '''
        if key !='':
            file.write('['+str(key)+'] '+str(value) + '\n')
        else:
            file.write('\n')

    def write_snap_metadata(self, path):
        metadata_path = os.path.dirname(path) + '/' + os.path.basename(path) + '_meta.txt'
        with open(metadata_path, 'w') as file:
            self.write_line(file, 'CFG')
            self.write_line(file, 'Laser', self.state['laser'])
            self.write_line(file, 'Intensity (%)', self.state['intensity'])
            self.write_line(file, 'Zoom', self.state['zoom'])
            self.write_line(file, 'Pixelsize in um', self.state['pixelsize'])
            self.write_line(file, 'Filter', self.state['filter'])
            self.write_line(file, 'Shutter', self.state['shutterconfig'])
            self.write_line(file)
            self.write_line(file, 'POSITION')
            self.write_line(file, 'x_pos', self.state['position']['x_pos'])
            self.write_line(file, 'y_pos', self.state['position']['y_pos'])
            self.write_line(file, 'z_pos', self.state['position']['z_pos'])
            self.write_line(file, 'f_pos', self.state['position']['f_pos'])
            self.write_line(file)
            ''' Attention: change to true ETL values ASAP '''
            self.write_line(file, 'ETL PARAMETERS')
            self.write_line(file, 'ETL CFG File', self.state['ETL_cfg_file'])
            self.write_line(file, 'etl_l_offset', self.state['etl_l_offset'])
            self.write_line(file, 'etl_l_amplitude', self.state['etl_l_amplitude'])
            self.write_line(file, 'etl_r_offset', self.state['etl_r_offset'])
            self.write_line(file, 'etl_r_amplitude', self.state['etl_r_amplitude'])
            self.write_line(file)
            self.write_line(file, 'GALVO PARAMETERS')
            self.write_line(file, 'galvo_l_frequency', self.state['galvo_l_frequency'])
            self.write_line(file, 'galvo_l_amplitude', self.state['galvo_l_amplitude'])
            self.write_line(file, 'galvo_l_offset', self.state['galvo_l_offset'])
            self.write_line(file, 'galvo_r_amplitude', self.state['galvo_r_amplitude'])
            self.write_line(file, 'galvo_r_offset', self.state['galvo_r_offset'])
            self.write_line(file)
            self.write_line(file, 'CAMERA PARAMETERS')
            self.write_line(file, 'camera_type', self.cfg.camera)
            self.write_line(file, 'camera_exposure', self.state['camera_exposure_time'])
            self.write_line(file, 'camera_line_interval', self.state['camera_line_interval'])
            self.write_line(file, 'x_pixels', self.cfg.camera_parameters['x_pixels'])
            self.write_line(file, 'y_pixels', self.cfg.camera_parameters['y_pixels'])

    @QtCore.pyqtSlot(Acquisition, AcquisitionList)
    def write_metadata(self, acq, acq_list):
        ''' Writes a metadata.txt file. Path contains the file to be written '''
        path = acq['folder'] + '/' + acq['filename']
        metadata_path = os.path.dirname(path) + '/' + os.path.basename(path) + '_meta.txt'

        if acq['filename'][-3:] == '.h5':
            if acq == acq_list[0]:
                self.metadata_file = open(metadata_path, 'w')
        else:
            self.metadata_file = open(metadata_path, 'w')

        self.write_line(self.metadata_file, 'Metadata for file', path)
        self.write_line(self.metadata_file, 'z_stepsize', acq['z_step'])
        self.write_line(self.metadata_file, 'z_planes', acq['planes'])
        self.write_line(self.metadata_file)
        # self.write_line(file, 'COMMENTS')
        # self.write_line(file, 'Comment: ', acq(['comment']))
        # self.write_line(file)
        self.write_line(self.metadata_file, 'CFG')
        self.write_line(self.metadata_file, 'Laser', acq['laser'])
        self.write_line(self.metadata_file, 'Intensity (%)', acq['intensity'])
        self.write_line(self.metadata_file, 'Zoom', acq['zoom'])
        self.write_line(self.metadata_file, 'Pixelsize in um', self.state['pixelsize'])
        self.write_line(self.metadata_file, 'Filter', acq['filter'])
        self.write_line(self.metadata_file, 'Shutter', acq['shutterconfig'])
        self.write_line(self.metadata_file)
        self.write_line(self.metadata_file, 'POSITION')
        self.write_line(self.metadata_file, 'x_pos', acq['x_pos'])
        self.write_line(self.metadata_file, 'y_pos', acq['y_pos'])
        self.write_line(self.metadata_file, 'f_start', acq['f_start'])
        self.write_line(self.metadata_file, 'f_end', acq['f_end'])
        self.write_line(self.metadata_file, 'z_start', acq['z_start'])
        self.write_line(self.metadata_file, 'z_end', acq['z_end'])
        self.write_line(self.metadata_file, 'z_stepsize', acq['z_step'])
        self.write_line(self.metadata_file, 'z_planes', acq.get_image_count())
        self.write_line(self.metadata_file, 'rot', acq['rot'])
        self.write_line(self.metadata_file)

        ''' Attention: change to true ETL values ASAP '''
        self.write_line(self.metadata_file, 'ETL PARAMETERS')
        self.write_line(self.metadata_file, 'ETL CFG File', self.state['ETL_cfg_file'])
        self.write_line(self.metadata_file, 'etl_l_offset', self.state['etl_l_offset'])
        self.write_line(self.metadata_file, 'etl_l_amplitude', self.state['etl_l_amplitude'])
        self.write_line(self.metadata_file, 'etl_r_offset', self.state['etl_r_offset'])
        self.write_line(self.metadata_file, 'etl_r_amplitude', self.state['etl_r_amplitude'])
        self.write_line(self.metadata_file)
        self.write_line(self.metadata_file, 'GALVO PARAMETERS')
        self.write_line(self.metadata_file, 'galvo_l_frequency', self.state['galvo_l_frequency'])
        self.write_line(self.metadata_file, 'galvo_l_amplitude', self.state['galvo_l_amplitude'])
        self.write_line(self.metadata_file, 'galvo_l_offset', self.state['galvo_l_offset'])
        self.write_line(self.metadata_file, 'galvo_r_amplitude', self.state['galvo_r_amplitude'])
        self.write_line(self.metadata_file, 'galvo_r_offset', self.state['galvo_r_offset'])
        self.write_line(self.metadata_file)
        self.write_line(self.metadata_file, 'CAMERA PARAMETERS')
        self.write_line(self.metadata_file, 'camera_type', self.cfg.camera)
        self.write_line(self.metadata_file, 'camera_exposure', self.state['camera_exposure_time'])
        self.write_line(self.metadata_file, 'camera_line_interval', self.state['camera_line_interval'])
        self.write_line(self.metadata_file, 'x_pixels', self.cfg.camera_parameters['x_pixels'])
        self.write_line(self.metadata_file, 'y_pixels', self.cfg.camera_parameters['y_pixels'])

        if acq['filename'][-3:] == '.h5':
            if acq == acq_list[-1]:
                self.metadata_file.close()
        else:
            self.metadata_file.close()

    def append_timing_info_to_metadata(self, acq, **kwargs):
        '''
        Appends a metadata.txt file
        Arguments:
            acq_start, img_start, img_end, acq_end: strings that contain acquisition and camera start/stop time.
            img_total_time: time count of taking images.
        '''
        for key in kwargs.keys():
            assert key in ('acq_start', 'img_start', 'img_end', 'acq_end', 'img_total_time'), f"Key unknown: {key}"
        path = acq['folder']+'/'+acq['filename']
        metadata_path = os.path.dirname(path)+'/'+os.path.basename(path)+'_meta.txt'
        with open(metadata_path,'a') as file:
            ''' Adding troubleshooting information '''
            self.write_line(file)
            self.write_line(file, 'TIMING INFORMATION')
            self.write_line(file, 'Started stack', kwargs.get('acq_start'))
            self.write_line(file, 'Started taking images', kwargs.get('img_start'))
            self.write_line(file, 'Stopped taking images', kwargs.get('img_end'))
            self.write_line(file, 'Stopped stack', kwargs.get('acq_end'))
            self.write_line(file, 'Frame rate:', str(acq.get_image_count()/kwargs.get('img_total_time')))



import sys
import pyqtgraph as pg
import numpy as np
import matplotlib
from PyQt5 import QtWidgets, QtCore
from PyQt5.uic import loadUi
import logging
logger = logging.getLogger(__name__)


class mesoSPIM_ContrastWindow(QtWidgets.QWidget):

    def __init__(self, parent=None):
        '''Parent can be a mesoSPIM_MainWindow() object'''
        super().__init__()
        self.parent = parent
        self.active = True
        self.image_view = pg.ImageView()
        cmap = pg.colormap.get('jet', source='matplotlib')
        self.image_view.setColorMap(cmap)
        self.setWindowTitle("Contrast map")
        layout = pg.QtGui.QVBoxLayout()
        self.setLayout(layout)
        layout.addWidget(self.image_view)
        self.resize(900, 900)
        self.show()

    def _contrast(self, roi):
        """Compute the contrast value, (max-min)/(max+min), from the image roi"""
        mini = np.percentile(roi, 1)
        maxi = np.percentile(roi, 99)
        contrast = (maxi - mini) / (maxi + mini)
        return contrast

    @QtCore.pyqtSlot(np.ndarray)
    def set_image(self, image):
        if self.active: # do updates only if widget window is active (visible), to minimize overhead.
            roi_h = roi_w = 128 # px, roi size
            N_ROIs_H, N_ROIs_W = int(np.ceil(image.shape[0] / roi_h)), int(np.ceil(image.shape[1] / roi_w))
            contrast_map = np.zeros((N_ROIs_H, N_ROIs_W))
            for j in range(N_ROIs_H):
                for i in range(N_ROIs_W):
                    roi = image[j * roi_h:(j + 1) * roi_h, i * roi_w:(i + 1) * roi_w]
                    contrast_map[j, i] = self._contrast(roi)
            self.image_view.setImage(contrast_map, autoLevels=False, autoHistogramRange=False, autoRange=False)

    def closeEvent(self, event):
        """Override close event: window becomes hidden and inactive, but still exists and ready to pop up when called"""
        logger.info("Image contrast window closed")
        self.active = False
        event.accept() # let the window close


if __name__ == '__main__':
    app = QtWidgets.QApplication([])
    window = mesoSPIM_ContrastWindow()
    sys.exit(app.exec_())



'''
Contains Image Processing Wizard Class: Sets online processing options

'''

from PyQt5 import QtWidgets, QtGui, QtCore

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import pyqtProperty

from ..mesoSPIM_State import mesoSPIM_StateSingleton

class ImageProcessingWizard(QtWidgets.QWizard):
    '''
    Wizard to run

    The parent is the Window class of the microscope
    '''
    wizard_done = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)

        ''' By an instance variable, callbacks to window signals can be handed
        through '''
        self.parent = parent
        self.cfg = parent.cfg
        self.state = mesoSPIM_StateSingleton()
  
        self.setWindowTitle('Image Processing Wizard')

        self.addPage(ImageProcessingWizardWelcomePage(self))
        self.addPage(ImageProcessingWizardSetOptionsPage(self))
        self.addPage(ImageProcessingWizardCheckResultsPage(self))
        
        self.show()

    def done(self, r):
        ''' Reimplementation of the done function

        if r == 0: canceled
        if r == 1: finished properly
        '''
        if r == 0:
            print("Wizard was canceled")
        if r == 1:
            print('Wizard was closed properly')
            # print('Laser selected: ', self.field('Laser'))
            self.set_processing_options()
        else:
            print('Wizard provided return code: ', r)

        super().done(r)

    def set_processing_options(self):
        row_count = self.parent.model.rowCount()
        processing_column = self.parent.model.getColumnByName('Processing')
         
        for row in range(0, row_count):
            index = self.parent.model.createIndex(row, processing_column)
            if self.field('maxProjEnabled'):
                self.parent.model.setData(index, 'MAX')
            else:
                self.parent.model.setData(index,  '')

class ImageProcessingWizardWelcomePage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle("Welcome to the image processing wizard!")
        self.setSubTitle("This wizard allows you to set image processing flags!")
    
class ImageProcessingWizardSetOptionsPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle("Select processing options")
        #self.setSubTitle("Select the processing options:")

        self.maxProjectionCheckBox = QtWidgets.QCheckBox('Post-stack MAX projection', self)
        
        self.registerField('maxProjEnabled', self.maxProjectionCheckBox)

        self.layout = QtWidgets.QGridLayout()
        self.layout.addWidget(self.maxProjectionCheckBox, 0, 0)
        self.setLayout(self.layout)

    def validatePage(self):
        '''Further validation operations can be introduced here'''
        return super().validatePage()

class ImageProcessingWizardCheckResultsPage(QtWidgets.QWizardPage):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.parent = parent

        self.setTitle('Options have been set')
        


        
    


'''
mesoSPIM State class
'''
import numpy as np
from PyQt5 import QtCore

from .utils.acquisitions import AcquisitionList

class mesoSPIM_StateSingleton():
    '''
    Singleton object containing the whole mesoSPIM state.

    Only classes which control.

    Access to attributes is mutex-locked to allow access from multiple threads.

    If more than one state parameter should be set at the same time, the 
    set_parameter 
    '''

    instance = None

    def __new__(cls):
        if not mesoSPIM_StateSingleton.instance:
            mesoSPIM_StateSingleton.instance = mesoSPIM_StateSingleton.__StateObject()

        return mesoSPIM_StateSingleton.instance

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def __setattr__(self, name):
        return setattr(self.instance, name)

    class __StateObject(QtCore.QObject):
        sig_updated = QtCore.pyqtSignal()
        mutex = QtCore.QMutex()

        def __init__(self):
            super().__init__()
            self._state_dict = {
                            'state' : 'init', # 'init', 'idle' , 'live', 'snap', 'running_script'
                            'acq_list' : AcquisitionList(),
                            'selected_row': -2,
                            'samplerate' : 100000,
                            'sweeptime' : 0.2,
                            'position' : {'x_pos':0,'y_pos':0,'z_pos':0,'f_pos':0,'theta_pos':0},
                            'ttl_movement_enabled_during_acq' : False,
                            'ETL_cfg_file' : 'config/etl_parameters/ETL-parameters.csv',
                            'filename' : 'file.tif',
                            'folder' : 'tmp',
                            'snap_folder' : 'tmp',
                            'file_prefix' : '',
                            'file_suffix' : '000001',
                            'zoom' : '1x',
                            'pixelsize' : 1.0,
                            'laser' : '488 nm',
                            'max_laser_voltage':1,
                            'intensity' : 10,
                            'shutterstate':False, # Is the shutter open or not?
                            'shutterconfig':'Right', # Can be "Left", "Right","Both","Interleaved"
                            'laser_interleaving':False,
                            'filter' : 'Empty-Alignment',
                            'etl_l_delay_%' : 7.5,
                            'etl_l_ramp_rising_%' : 85,
                            'etl_l_ramp_falling_%' : 2.5,
                            'etl_l_amplitude' : 0.7,
                            'etl_l_offset' : 2.3,
                            'etl_r_delay_%' : 2.5,
                            'etl_r_ramp_rising_%' : 5,
                            'etl_r_ramp_falling_%' : 85,
                            'etl_r_amplitude' : 0.65,
                            'etl_r_offset' : 2.36,
                            'galvo_l_frequency' : 99.9,
                            'galvo_l_amplitude' : 6,
                            'galvo_l_offset' : 0,
                            'galvo_l_duty_cycle' : 50,
                            'galvo_l_phase' : np.pi/2,
                            'galvo_r_frequency' : 99.9,
                            'galvo_r_amplitude' : 6,
                            'galvo_r_offset' : 0,
                            'galvo_r_duty_cycle' : 50,
                            'galvo_r_phase' : np.pi/2,
                            'laser_l_delay_%' : 10,
                            'laser_l_pulse_%' : 87,
                            'laser_l_max_amplitude_%' : 100,
                            'laser_r_delay_%' : 10,
                            'laser_r_pulse_%' : 87,
                            'laser_r_max_amplitude_%' : 100,
                            'camera_delay_%' : 10,
                            'camera_pulse_%' : 1,
                            'camera_exposure_time':0.02,
                            'camera_line_interval':0.000075,
                            'camera_display_live_subsampling': 1,
                            'camera_display_acquisition_subsampling': 2,
                            'camera_binning':'1x1',
                            'camera_sensor_mode':'ASLM',
                            'current_framerate':3.8,
                            'predicted_acq_list_time':1,
                            'remaining_acq_list_time':1,
                            }

        def __len__(self):
            return len(self._state_dict) 
        
        def __setitem__(self, key, value):
            '''
            Custom __setitem__ method to allow mutexed access to 
            a state parameter. 

            After the state has been changed, the updated signal is emitted.
            '''
            with QtCore.QMutexLocker(self.mutex):
                self._state_dict.__setitem__(key, value)
            self.sig_updated.emit()

        def __getitem__(self, key):
            '''
            Custom __getitem__ method to allow mutexed access to 
            a state parameter.

            To avoid the state being updated while a parameter is read.
            '''

            with QtCore.QMutexLocker(self.mutex):
                return self._state_dict.__getitem__(key)

        def set_parameters(self, dict):
            '''
            Sometimes, several parameters should be set at once 
            without allowing the state being updated while a parameter is read.
            '''
            with QtCore.QMutexLocker(self.mutex):
                for key, value in dict.items():
                    self._state_dict.__setitem__(key, value)
            self.sig_updated.emit()

        def get_parameter_dict(self, list):
            '''
            For a list of keys, get a state dict with the current values back.

            All the values are read out under a QMutexLocker so that 
            the state cannot be updated at the same time.
            '''
            return_dict = {}

            with QtCore.QMutexLocker(self.mutex):
                for key in list:
                    return_dict[key] = self._state_dict.__getitem__(key)
            
            return return_dict

        def get_parameter_list(self, list):
            '''
            For a list of keys, get a state list with the current values back.

            This is especially useful for unpacking.

            All the values are read out under a QMutexLocker so that 
            the state cannot be updated at the same time.
            '''
            return_list = []

            with QtCore.QMutexLocker(self.mutex):
                for key in list:
                    return_list.append(self._state_dict.__getitem__(key))
            
            return return_list

        def block_signals(self, boolean):
            self.blockSignals(boolean)
        